/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/


package commandline;

import java.io.Reader;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.IOException;

/**
 * Some terminals do not handle backspaces gracefully, they
 * display '^H' for backspace for instance. This filter interprets
 * backspace characters, eliminating the characters of the input
 * that were erased by the backspaces.
 *
 * This filter cannot return to the caller until it reads an
 * entire line, otherwise it cannot know how many backspace
 * characters are present. Due to this restriction, this class
 * does not extend <code>FilterReader</code>.
 *
 * @see "The Java Programming Language, 3rd ed. p 372 section 15.4.2 Filter Streams"
 * @see BufferedReader
 *
 * @author HyperCast Team
 * @author Greg Mattes, October 2004
 */
public class EliminateBackspaceFilter
    extends BufferedReader
{
    private String lineBuffer = null;
    private int lineBufferIndex = 0;

    public EliminateBackspaceFilter (final Reader in)
    {
        super (in);
    }

    /**
     * @see BufferedReader#readLine()
     * @see #readLineBuffer()
     */
    public String readLine()
        throws IOException
    {
        if (lineBuffer == null)
        {
            readLineBuffer();
        }

        String result = null;
        if (lineBuffer != null) // null -> end-of-stream
        {
            result = lineBuffer.substring (lineBufferIndex);
            lineBuffer = null;
            lineBufferIndex = 0;
        }
        return result;
    }

    /**
     * @see BufferedReader#read()
     * @see #readLineBuffer()
     */
    public int read()
        throws IOException
    {
        final int c;

        if (lineBuffer == null)
        {
            readLineBuffer();
        }

        if (lineBuffer == null) // end-of-stream
        {
            c = -1;
        }
        else
        {
            c = (int) lineBuffer.charAt (lineBufferIndex);
            ++lineBufferIndex;
            if (lineBufferIndex >= lineBuffer.length())
            {
                lineBuffer = null;
                lineBufferIndex = 0;
            }
        }
        return c;
    }

    /**
     * @see BufferedReader#read(char[],int,int)
     * @see #readLineBuffer()
     */
    public int read (final char [] buffer, final int offset, final int length)
        throws IOException
    {
        if (lineBuffer == null)
        {
            readLineBuffer();
        }

        final int result;
        if (lineBuffer == null) // end-of-stream
        {
            result = -1;
        }
        else
        {
            final int count = Math.min (length, lineBuffer.length());
            lineBuffer.getChars (0, count, buffer, offset);
            result = count;
        }

        return result;
    }

    /**
     * Populates the instance variable <code>lineBuffer</code> by
     * reading a line from the underlying stream. The line of
     * characters is purged of backspace characters.
     */
    private void readLineBuffer()
        throws IOException
    {
        if (lineBuffer != null) // do not need to read a line - already have one! no-op
        {
            return;
        }

        lineBufferIndex = 0;
        lineBuffer = super.readLine();

        if (lineBuffer == null) // end-of-stream, handled by caller
        {
            return;
        }
               
        lineBuffer = eliminateBackspaceCharacters (lineBuffer);
    }

    /**
     * Finds the first and last occurrence of every sequence of
     * backspace characters. Removes such sequences as well as the
     * characters before the sequence that should be erased.
     *
     * @param s A string that might contain backspace characters
     *
     * @return A string with no backspace characters
     */
    private String eliminateBackspaceCharacters (final String s)
    {
        final StringBuffer buf = new StringBuffer (s);

        int firstBackspace;
        while ((firstBackspace = buf.indexOf ("\b")) != -1)
        {
            int lastBackspace = firstBackspace;

            while (lastBackspace < buf.length()
                   &&
                   buf.charAt (lastBackspace) == '\b')
            {
                ++lastBackspace;
            }

            final int backspaceCount = lastBackspace - firstBackspace;
            final int start = Math.max (0, firstBackspace - backspaceCount); // inclusive, non-negative start
            final int end = firstBackspace + backspaceCount; // exclusive

            buf.delete (start, end);
        }

        return buf.toString();
    }

    public boolean markSupported()
    {
        return false;
    }

    public void mark (final int readAheadLimit)
    {
        throw new UnsupportedOperationException ("Mark operation is not supported by EliminateBackspaceFilter");
    }

    public void reset()
    {
        throw new UnsupportedOperationException ("Reset operation is not supported by EliminateBackspaceFilter");
    }

    public long skip (final long n)
    {
        throw new UnsupportedOperationException ("Skip operation is not supported by EliminateBackspaceFilter");
    }

    /**
     * Test driver
     */
    public static void main (final String [] args)
    {
        try
        {
            final char [] buf = new char [20];
            final BufferedReader br = new EliminateBackspaceFilter (new InputStreamReader (System.in));
            final int c = br.read();
            System.out.println ("char: " + (char)c);
            final String l = br.readLine();
            System.out.println ("line: " + l);
            final int i = br.read (buf, 0, 20);
            System.out.println ("char read: " + i + " buffer: " + (new String (buf, 0, i)));
        }
        catch (IOException ioe)
        {
            System.err.println ("I/O Exception: " + ioe);
        }
    }
}
