/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

/* 
 * @author HyperCast Team
 */

package commandline;

import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.io.IOException;

import java.util.Date;

import runcontrol.RunControl;
import runcontrol.CommandFailureException;

public class RunControlCLI
{
    private static final String DEFAULT_CONFIGURATION_FILENAME = "RunControlSocket.xml";
    private static final String PROMPT = "> ";
	
    /** Command processor that has a monitor socket for use with Monitor and Control protocol */
    private RunControl runControl;

    /**
     * Determines configuration file name for monitor socket and
     * instanciates a command processor object.
     */
    public RunControlCLI (final String args[])
    {
        this.runControl = createRunControl(args, "commandline.RunControlCLI");
    }

    
    /**
     * Displays the valid forms of command line interface invocation.
     */
    private static void usage(String classname)
    {
        final String pathSeparator = System.getProperty ("path.separator");
        System.err.println ("usage: java -cp runcontrol.jar" + pathSeparator + "hypercast.jar " +
        					classname + " " +
                            "[-mc Monitor Configuration Filename] " +
                            "[-qr Maximum times to retry Query after timeouts] " +
                            "\n");
        System.exit (1);
    }

    /**
     * Prepares input/output readers and writers for use with the
     * command line interpreter.
     */
    private void cliEntryPoint()
    { 
        final BufferedReader stdin = new EliminateBackspaceFilter (new InputStreamReader (System.in));
        final PrintWriter stdout = new PrintWriter (new OutputStreamWriter (System.out), true);
        final PrintWriter stderr = new PrintWriter (new OutputStreamWriter (System.err), true);
        commandLine (stdin, stdout, stderr);
    }

    /**
     * Receives commands on the InputStream and processes them.
     * This is connected to the console by go().  A command could connect it to 
     * a TCP socket for remote access.   This function catches all exceptions.  
     * If the exception is Exception("quit"), then it will halt; all other exceptions will 
     * cause the stack to be printed and then operation will continue.
     */
    private void commandLine (BufferedReader in, PrintWriter out, PrintWriter err)
    {
        out.println ("Started RunControl Command-line interpreter at " + (new Date()));

        boolean continueRunning = true;

        while (continueRunning)
        {
            try
            {
                out.print (PROMPT);
                out.flush();

                final String input = in.readLine();
                if (input == null) // end-of-stream encountered
                {
                    out.println();
                    continueRunning = false;
                }
                else if (! input.trim().equals (""))
                {
                    //parsedCommand = runControl.parseInput (input); // TODO
                    /*continueRunning =*/ runControl.processCommand (input, in, out, err);
                }
            }
            catch (final IOException ioe)
            {
                err.println ("I/O Exception occured: " + ioe);
                ioe.printStackTrace (err);
                err.println ("Assuming transient failure, continuing...");
            }
            catch (final CommandFailureException cfe)
            {
                out.println (cfe.getMessage());
            }
            catch (final Exception e)
            {
            	if (e.getMessage() != null
                        &&
                        e.getMessage().toUpperCase().trim().equals ("QUIT"))
                {
            		try{
            			runControl.getMonitor().closeMonitor();
            		}finally{
                        System.exit(0);
            		}
                }
            	
                err.println ("Encountered exception that cannot be handled: " + e);
                e.printStackTrace (err);
                try{
                	runControl.getMonitor().closeMonitor();
                }finally{
                	continueRunning = false;
                }
            }
        }
    }
    public static RunControl createRunControl(String [] args, String classname){
    	String monitorConfigurationFilename = DEFAULT_CONFIGURATION_FILENAME;
        int maximumQueryRetry = 3;
        
        for (int x = 0; x < args.length; ++x)
        {
            if (args[x].equals ("-mc"))
            {
                ++x;
                if (args.length < x+1)
                {
                    usage(classname);
                }
                monitorConfigurationFilename = args[x];
            }
            else if (args[x].startsWith ("-qr"))
            {
                ++x;
                if (args.length < x+1)
                {
                    usage(classname);
                }
                
                try
                {
                    maximumQueryRetry = Integer.parseInt (args[x]);
                    if (maximumQueryRetry <= 0)
                        throw new NumberFormatException();
                }
                catch (final NumberFormatException nfe)
                {
                    System.err.println ("Bad argument for number of maximum query retries: " + args[x]);
                    usage(classname);
                }
            }
            // These options are superfluous since anything that is
            // not recognized does the same thing.
            else if (args[x].equals ("-h")
                     ||
                     args[x].equals ("-help"))
            {
                usage(classname);
            }
            else
            {
                usage(classname);
            }
        }
        return new RunControl(monitorConfigurationFilename, maximumQueryRetry);
    	
    }
    /**
     * Command line interface entry point.
     *
     * @see #usage
     */
    public static void main (final String [] args)
    {
        

        RunControlCLI runControlCLI = new RunControlCLI (args);
        runControlCLI.cliEntryPoint();

        System.exit (0);
    }
}
