/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

/* 
 * @author HyperCast Team
 * July, 2002 
 */

package gui;

import java.lang.*;
import java.awt.*;
import java.util.*;

/**
 * Overlay Node, composed of physical and logical parts
 * Main components </p>
 * a. physical address </p>
 * b. physical links </p>
 * c. logical links </p>
 * d. physical name </p>
 * e. routing table to provide the physical network service </p>
 * f. protocol(logical part) </p>
 * Main functions </p>
 * a. maintenence of the physical/logical links </p>
 *      note: the physical links are added/deleted by outside
 *            the logical links are added/deleted by the protocol/graph-sim
 * b. dispatch/receive the events for two layers. </p>
 *          interface to the event dispatcher: send/receive the event(timer/msg)
 *          interface to the protocol: send/receive the event(timer/msg)
 *          for the message, it may need to route the message to the next hop
 * c. add/delete/start/end the running protocol </p>
 */

class OverlayNode implements physicalInterface{

    final static float PACKET_SPEED = 1000.0f;
    //node attributes
    /** name of the node */
    String name;
     /** physical node*/
     PhyNode phyNode;
    /** physical address of the node */
    XYAddress addr;

    //logical structure
    /**
     * protocol running on this node. It can be empty
     * if no protocol is running
     */
    protocolInterface protocol;
    /** neighbors in the overlay */
    Vector logical_neighbors;      // logical neighbors
    /** color used to display edges in the overlay */
    Vector logical_colors;         // the link color, display purpose
    /** the width of the logical edges */
    Vector logical_width;          // the link width, display purpose;

    //physical structure
    /** physical neighbors */
    Vector physical_neighbors;       // physical neighbors
    /** the lenght of the physical edges */
    Vector physical_distances;       // physical neighbors

    //routing table, the next hop is indexed by the node's position
    //in the RunNetwork's nodes vector
    /** routing table at this node. It stores the next hope node from this
     *  node to all the other nodes
     *
     */
    OverlayNode [] routing_nexthop;
    /** a referece to the class Overlay Network */
    RunNetwork net;
    
     /** time when the last message sent by this node will be delivered (junk data) */
     public float timeLastMsg;

    /**constructor
     * @param a node's physical address
     * @param onet a reference to the overlay network
     * @param sname node's name
     */
    public OverlayNode(XYAddress a, RunNetwork onet, String sname)
    {
        phyNode = new PhyNode(a.getX(), a.getY());
        addr = a;
        net = onet;
        name = sname;
        logical_neighbors = new Vector();
        logical_colors = new Vector();
        logical_width = new Vector();
        physical_neighbors = new Vector();
        physical_distances = new Vector();
        
    }

    /**
     * @param pnode node's physical address
     * @param onet a reference to the overlay network
     * @param sname node's name
     */
    public OverlayNode(PhyNode pnode, RunNetwork onet, String sname)
    {
        phyNode = pnode;
        addr = new XYAddress(pnode.GetX(),pnode.GetY());
        net = onet;
        name = sname;
        logical_neighbors = new Vector();
        logical_colors = new Vector();
        logical_width = new Vector();
        physical_neighbors = new Vector();
        physical_distances = new Vector();
        
    }
    
    
    //attributes method
    /** @return node name */
    public String getName() { return name; }
    public String getDetails() {
        if(protocol!=null){
            return protocol.getDetails();
        }else return "";    
    }
    
    /** set the name of this node */
    public void setName( String sname) { name = sname; }
    /** @return the physical address of the node */
    public XYAddress getAddress()  {   return new XYAddress(phyNode.GetX(),phyNode.GetY()); }
    /** set the address of the node */
    public void setAddress( XYAddress a ) { 
        phyNode.SetXF(a.getX());
        phyNode.SetYF(a.getY());
    }

    //logical & physical links' related method
    /**
     * test if the node is its physical neighbor
     */
    boolean hasPhyEdge( OverlayNode n ) {
        if( physical_neighbors.contains(n) ) return true;
        else return false;
    }

    /**
     * test if the node is its logical neighbor
     */
    boolean hasLogicalEdge( OverlayNode n){
        if( logical_neighbors.contains(n)) return true;
        else return false;
    }

    /**
     * insert a physical edge for this node
     * @param n the new neigbhor
     * @param weight the length of the node
     */
    boolean insertPhyEdge( OverlayNode n, float weight ){
        if( physical_neighbors.contains(n) ) return false;
        else  {
            physical_neighbors.addElement(n);
            physical_distances.addElement( new Float(weight) );
        }
        return true;
    }

    /**
     * @return all the physical neighbors of this node
     */
    Enumeration getPhyNeighbor(){
        return physical_neighbors.elements();
    }

    /**
     * @return all the logical neighbors of this node
     */
    Enumeration getLogicalNeighbor(){
        return logical_neighbors.elements();
    }

    /**
     * @return the colors of the logical links
     */
    Enumeration getLogicalColor() {
        return logical_colors.elements();
    }

    /**
     * @return the lengths of the logical links
     */
    Enumeration getLogicalWidth() {
        return logical_width.elements();
    }

    /**
     * @return the total number of logical neighbors
     */
    int getLogicalNeighborNumber(){ return logical_neighbors.size();}
    /**
     * @return the total number of physical neighbors
     */
    int getPhyNeighborNumber( ) { return physical_neighbors.size(); }
    /**
     * @return the index of this physical node in all the neighbors
     */
    int getPhyNeighborIndex( OverlayNode onode ) {
        return physical_neighbors.indexOf( onode );
    }

    /**
     * @return the legth of a physical edge, which is represented by
     * its index among all the neighbors
     */
    float getPhyEdgeLen( int index ){
        return ((Float)physical_distances.elementAt(index)).floatValue();
    }

    /**
     * remove an physical edge
     * @param n the node at the other end of the edge
     * @return true if such a neighbor exists, false otherwise
     */
    boolean removePhyEdge( OverlayNode n ){
        if( physical_neighbors.contains(n) ) {
            int index = physical_neighbors.indexOf(n);
            physical_neighbors.remove(index);
            physical_distances.remove(index);
            return true;
        }else{
            return false;
        }
    }


    /**
     * clear all the logical edges
     */
    public void clearLogicalEdges(){
        this.logical_colors.clear();
        this.logical_neighbors.clear();
        this.logical_width.clear();
    }

    //physicalInterface
    /**
     * part of the physicalInterface implemenation
     * @return the logical neighbors of this node
     */
    public Enumeration logicalLinks() {
        return getLogicalNeighbor();
    }

    
    /** <0 if the physical node is not a neighbor
     *  otherwise, the link weight
     */
    //float physicalWeight( physicalInterface );

    /**
     * @return the physical address of this node
     */
    public XYAddress physicalAddress(){
        return getAddress();
    }

    /**
     * insert a logical edge at this node
     * @param n the other end of the logical edge
     * @param color the color of the edge, for display purpose
     * @param width the width of the edge, for dispaly purpose
     */
    public boolean insertLogicalEdge( OverlayNode n, Color color, int width ){
        if( logical_neighbors.contains(n) ) return false;
        else  {
            logical_neighbors.addElement(n);
            logical_colors.addElement(color);
            logical_width.addElement( new Integer(width) );
        }
        return true;
    }

    /**
     * remove a logical neighbor
     */
    public boolean removeLogicalEdge(OverlayNode n){
        if( logical_neighbors.contains(n) ){
            int index = logical_neighbors.indexOf(n);
            logical_neighbors.remove(index);
            logical_colors.remove(index);
            logical_width.remove(index);
            return true;
        }
        else{
            return false;
        }
    }


    /** set the logical node running on this node */
    void setProtocol( protocolInterface p ) {
        protocol = p;
    }

    /** return the logical node running on this node*/
    protocolInterface getProtocol( ) {
        return protocol;
    }
}
