/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/


package runcontrol;

import hypercast.MonitorAndControl.I_ReceiveCallback;
import hypercast.MonitorAndControl.I_TimeoutCallback;
import hypercast.MonitorAndControl.I_Monitor;
import hypercast.MonitorAndControl.MonMessage;
import hypercast.MonitorAndControl.NoSuchPortalException;

/**
 * <p>This class provide a way to send out reliable MonMessage to
 * Portal. Sending process can be blocked until a reply is
 * received. RunControl program can use this reliable sending
 * function to support the running of a command.</p>
 * 
 * @author HyperCast Team
 * @author Guangyu Dong, Sept 24, 2004
 * @author Greg Mattes, October, 2004
 */
public class CommandRunner
    implements I_ReceiveCallback, I_TimeoutCallback
{
    /** The <code>Monitor</code> instance that will send queries, receive responses, and handle timeouts on behalf of this <code>CommandRunner</code>. */
    private I_Monitor monitor;

    /** The number of times a message will be resent if timeouts are occurring. */
    private int maximumQueryResends;

    /** Contains the messages that comprise the current set of pending queries. If <code>null</code> there is no query pending. */
    private MonMessage [] queryMessages = null;
    
    /** Contains the index of portals where the messages are sent to */
    private int [] portals = null;

    /** Contains reply messages corresponding to query messages; <code>null</code> entries indicate timeouts for corresponding queries. Corresponds to the <code>queryMessages</code> array.*/
    private MonMessage [] replyMessages = null;

    /** Contains the number of times for each message that a resend has occurred. Corresponds to the <code>queryMessages</code> array. */
    private int [] resendTracker = null;

    /** Used to keep track of how many replies have been received. The number of pending queries is <code>queryMessage.length - replyCount - timeoutCount</code>. */
    private int replyCount = 0;

    /** Used to keep track of how many timeouts have been detected. The number of pending queries is <code>queryMessage.length - replyCount - timeoutCount</code>. */
    private int timeoutCount = 0;

    /** Thread that called <code>sendQuery</code>/<code>sendQueries</code>. If <code>null</code> there is no query pending. */
    private Thread queryThread = null;

    /**
     * <p>Initializes a <code>CommandRunner</code> object.</p>
     *
     * @param monitor The monitor.
     * @param timeout Timeout time.
     * @param maximumQueryResends Maximum resends for any message before timeout
     */
    public CommandRunner (final I_Monitor monitor, final int maximumQueryResends)
    {
        this.monitor = monitor;
        this.maximumQueryResends = maximumQueryResends;
    }

    
    
    public synchronized void setMaximumQueryResends (final int maximumQueryResends)
        throws InterruptedException
    {
        while (queryMessages != null)
        {
            wait(); // wait until current query finishes
        }

        this.maximumQueryResends = maximumQueryResends;
    }

    public synchronized int getMaximumQueryResends()
    {
        return maximumQueryResends;
    }

    /**
     * <p>Sends a message without waiting for reply, i.e. non-blocking.</p>
     *
     * NB: With the current implementation there is no way to receive
     * a reply to an unblocked query using this interface.
     *
     * @param portalIndex Index of portal (TODO: what defines valid values?).
     * @param query Message containing query that is to be sent to a remote portal.
     *
     * @throws NoSuchPortalException
     */
    public void sendUnblockedQuery (final int portalIndex, final MonMessage query)
        throws NoSuchPortalException
    {
        monitor.sendTo (portalIndex, query, null, null);
    }

    /**
     * <p>Sends a batch messages without waiting for reply,
     * i.e. non-blocking. This method is nothing more than a thin
     * wrapper around <code>sendUnblockedQuery()</code>.</p>
     *
     * NB: With the current implementation there is no way to receive
     * a reply to an unblocked query using this interface.
     *
     * @see #sendUnblockedQuery(int,MonMessage)
     *
     * @param portalIndices A list of indices of portals (TODO: what defines valid values?).
     * @param queries Messages containing queries that are to be sent to a remote portals.
     *
     * @throws NoSuchPortalException
     */
    public void sendUnblockedQueries (final int [] portalIndices, final MonMessage [] queries)
        throws NoSuchPortalException
    {
        for (int i = 0; i < queries.length; ++i)
        {
            sendUnblockedQuery (portalIndices[i], queries[i]);
        }
    }

    /**
     * <p>Sends a message to all portals without waiting for reply,
     * i.e. non-blocking.</p>
     *
     * NB: With the current implementation there is no way to receive
     * a reply to an unblocked query using this interface.
     *
     * @param query Message containing query that is to be sent to a remote portal.
     */
    public void sendUnblockedQueryToAll (final MonMessage query)
    {
        monitor.sendToAll (query, null, null);
    }

    /**
     * <p>Sends a batch of messages to all portals without waiting for
     * reply, i.e. non-blocking. This method is nothing more than a
     * thin wrapper around <code>sendUnblockedQueryToAll()</code>.</p>
     *
     * NB: With the current implementation there is no way to receive
     * a reply to an unblocked query using this interface.
     *
     * @see #sendUnblockedQueryToAll(MonMessage)
     *
     * @param queries Messages containing queries that are to be sent to a remote portals.
     */
    public void sendUnblockedQueriesToAll (final MonMessage [] queries)
    {
        for (int i = 0; i < queries.length; ++i)
        {
            sendUnblockedQueryToAll (queries[i]);
        }
    }

    /**
     * <p>Sends a message and waits for a reply, i.e. blocking. Can be
     * interrupted; interruption is distinct from timeout.</p>
     *
     * @param portalIndex Index of portal (TODO: what defines valid values?).
     * @param query The message that contains the query.
     *
     * @see #sendQueries(int[],MonMessage[])
     *
     * @return The query reply; can be <code>null</code> if interrupted or timeout.
     *
     * @throws InterruptedException If query is interrupted.
     * @throws NoSuchPortalException If <code>portalIndex</code> does not specify a valid Portal
     */
    public MonMessage sendQuery (final int portalIndex, final MonMessage query)
        throws InterruptedException, NoSuchPortalException
    {
        final int [] portalIndices = new int [] { portalIndex };
        final MonMessage [] queries = new MonMessage [] { query };
        final MonMessage [] replies = sendQueries (portalIndices, queries);
        return replies[0];
    }

    /**
     * <p>Sends messages and waits for replies,
     * i.e. blocking. Provides a blocking service that is implemented
     * in terms of the non-blocking service provided by
     * <code>Monitor</code>.</p>
     *
     * <p>This method is designed to be interrupted; interruption is
     * distinct from timeout. On interruption an
     * <code>InterruptedException</code> is throw; the entire
     * operation, regardless of the number of individual queries, is
     * interrupted. A timeout is relative to a particular query in the
     * set of queries. This method will execute successfully in the
     * presence of timeouts, the caller must scan for
     * <code>null</code> in the array returned.</p>
     *
     * <p>This method ensures that even though many threads invoke it,
     * the individual queries are executed in some serial order.</p>
     *
     * @param portalIndices Indices of portals.
     * @param queries The messages that contains the queries.
     *
     * @return An array of reply messages. Entries having value of
     * <code>null</code> indicate timeouts. The entries in the
     * returned array correspond to the messages in the
     * <code>queries</code> array; i.e. <code>replies[i]</code> is the
     * reply for <code>queries[i]</code> where <code>replies</code> is
     * the name by which the returned array is referenced.
     *
     * @throws InterruptedException If query is interrupted.
     *
     * @throws NoSuchPortalException If an invalid portal index is
     * given - some messages may have already be sent when this
     * exception is thrown.
     */
    public MonMessage [] sendQueries (final int [] portalIndices, final MonMessage [] queries)
        throws InterruptedException, NoSuchPortalException
    {
        synchronized (this)
        {
            while (queryMessages != null)
            {
                wait(); // wait until current query finishes; NB: queries not served in FIFO order - order decided by thread scheduler.
            }

            queryThread = Thread.currentThread();
            queryMessages = queries;
            portals = portalIndices;

            replyCount = 0;
            timeoutCount = 0;
        
            // create space to store reply messages and count resends
            // *before* query messages are sent since replies can
            // arrive and timeouts can occur before the final query is
            // sent.

            resendTracker = new int [queries.length];
            replyMessages = new MonMessage [queries.length];
        }

        // unsynchronize so that receive and timeout methods can
        // execute while messages are being sent.

        for (int i = 0; i < queryMessages.length; ++i)
        {
            if (Thread.interrupted())
                throw new InterruptedException();

            monitor.sendTo (portalIndices[i], queryMessages[i], this, this); // non-blocking
        }

        synchronized (this)
        {
            try
            {
                /*
                 * wait for replies which will be received by the receiveMCMessage method
                 * and for timeouts which will be handled by the timeoutMCMessage method.
                 */
                while ((replyCount + timeoutCount) < queryMessages.length)
                {
                    // wait until query reply is received, or a timeout
                    // occurs.  Do not catch exception on interrupt,
                    // instead throw the exception, this allows the caller
                    // to interrupt a query.

                    wait();
                }

                return replyMessages;
            }
            finally
            {
                queryMessages = null;
                queryThread = null;
                replyCount = 0;
                timeoutCount = 0;
                notifyAll(); // allow waiting threads to submit query.
            }
        }
    }

    /**
     * <p>Processes a reply to a query. Matches the reply to a query, if
     * a match is found the reply is saved and the number of total
     * replies is incremented, finally other threads are signaled
     * about the reply arrival.</p>
     *
     * <p>When a blocking query is performed, the current instance of
     * <code>CommandRunner</code>, i.e. <code>this</code>, is passed
     * as a callback parameter. The <code>Monitor</code> that handles
     * the query will invoke this method if a reply is received for the
     * query.</p>
     *
     * @param reply The message that is the reply to a query.
     *
     * @see hypercast.MonitorAndControl.I_ReceiveCallback#receiveMessage(hypercast.MonitorAndControl.MonMessage)
     */
    synchronized public void receiveMessage (final MonMessage reply)
    {
        if (queryMessages == null)
        {
            return;
        }

        for (int i = 0; i < queryMessages.length; ++i)
        {
            if (reply.equals (queryMessages[i]))
            {
                // ignore dups TODO: console message TODO: should dups
                // be ignored or should timestamps be compared? could
                // choose the most recent message because it has the
                // freshest data.

                if (replyMessages[i] != null)
                {
                    System.err.println ("Notice: CommandRunner ignoring duplicate query response from Portal");
                }
                else
                {
                    replyMessages[i] = reply;
                    ++replyCount;
                    notifyAll(); // Signal receipt of reply message.
                }
                return;
            }
        }
    }

    /**
     * <p>Handles a timeout of a blocking query. Matches the timeout to a
     * query, if a match is found the number of total timeouts is
     * incremented, finally other threads are signaled about timeout.</p>
     *
     * <p>When a blocking query is done, the current instance of
     * <code>CommandRunner</code>, i.e. <code>this</code>, is passed
     * as a parameter. The <code>Monitor</code> that handles the call
     * will invoke this method if the query timeout time expires
     * before a reply is received.</p>
     *
     * @param timeoutMessage Query message that has timed out.
     *
     * @see hypercast.MonitorAndControl.I_Monitor
     * @see hypercast.MonitorAndControl.I_TimeoutCallback#timeoutMessage(hypercast.MonitorAndControl.MonMessage)
     */
    synchronized public void timeoutMessage (final MonMessage timeoutMessage)
    {
        if (queryMessages == null)
        {
            return;
        }

        boolean resend = false;
        for (int i = 0; i < queryMessages.length; ++i)
        {
            if (timeoutMessage.equals (queryMessages[i]))
            {
                if (resendTracker[i] < maximumQueryResends)
                {
                    ++resendTracker[i];
                    resend = true;
                    
                    /* resend is done here */
                    try {
						monitor.sendTo(portals[i], timeoutMessage, this, this);
					} catch (NoSuchPortalException e) {
						e.printStackTrace();
					}
                }
                else
                {
                    ++timeoutCount;
                    notifyAll(); // Signal timeout of query.
                    resend = false;
                }

                // debugging
                //System.err.println ("Timeout for QUERY: message id: " + queryMessages[i].getMessageID() + ", destination: " + queryMessages[i].getDestination() + ", resend: " + resend);

                break;
            }
        }
        
    }

    /**
     * <p>Interrupts a pending query if there is one.</p>
     */
    synchronized public void interruptQuery()
    {
        if (queryThread != null)
        {
            queryThread.interrupt();
        }
    }
}
