/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/


package runcontrol;

import hypercast.HyperCastConfig;
import hypercast.MonitorAndControl.I_Monitor;

import java.io.BufferedReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.Hashtable;
import java.util.StringTokenizer;

import runcontrol.commands.CheckClockSynchronizationCommand;
import runcontrol.commands.ClearPortalListCommand;
import runcontrol.commands.CompactPortalListCommand;
import runcontrol.commands.DateCommand;
import runcontrol.commands.EchoCommand;
import runcontrol.commands.FlushSocketsCommand;
import runcontrol.commands.GetExactStableTimeCommand;
import runcontrol.commands.GetExactStartTimeCommand;
import runcontrol.commands.GetValueCommand;
import runcontrol.commands.HelpCommand;
import runcontrol.commands.KillRemoteServersCommand;
import runcontrol.commands.ListCommand;
import runcontrol.commands.ListSocketsCommand;
import runcontrol.commands.MTimeCommand;
import runcontrol.commands.PauseCommand;
import runcontrol.commands.PollStabilityCommand;
import runcontrol.commands.QuitCommand;
import runcontrol.commands.RunScriptCommand;
import runcontrol.commands.SaveReadSchemaCommand;
import runcontrol.commands.SaveWriteSchemaCommand;
import runcontrol.commands.SetGridCoordsCommand;
import runcontrol.commands.SetNamesCommand;
import runcontrol.commands.SetPassiveModeCommand;
import runcontrol.commands.SetValueCommand;
import runcontrol.commands.SleepCommand;
import runcontrol.commands.StartSocketsCommand;
import runcontrol.commands.StartTranscriptCommand;
import runcontrol.commands.StopSocketsCommand;
import runcontrol.commands.StopTranscriptCommand;
import runcontrol.commands.TimeoutValuesCommand;
import runcontrol.commands.WaitUntilStableCommand;

/** 
 * @author HyperCast Team
 * @author Guangyu Dong
 * @author Greg Mattes
 * @author Konrad Lorincz
 * @author Michael Nahas
 * @version Spring 2005
 */

public class RunControl
{
    /** The maximum number of attempts to send a message before failure. */
    //public static final int NUM_OF_RETRIES_BEFORE_FAILURE = 20;
    public static final int NUM_OF_RETRIES_BEFORE_FAILURE_DEFAULT = 3;

    /**
     * Monitor object, this is an overlay socket that is used to
     * connect to a control network. A monitor communicates with one
     * or more portals.
     *
     * @see hypercast.MonitorAndControl.I_Monitor
     * @see hypercast.MonitorAndControl.I_Portal
     */
    private I_Monitor monitor;

    
    /** Socket Status Table */
    SocketStatusTable socketsStatusTable;
    
    /**
     * A Hashtable mapping the commands, e.g. "help" (represented as
     * Strings) to their associated RunControlCommand objects.
     */
    // TODO: make private?
    protected Hashtable commands = new Hashtable(); // String -> RunControlCommand
	
    /** current command runner */
    private CommandRunner commandRunner;
    
    /** config object */
    protected HyperCastConfig config;

    /**
     * Inserts generic RunControlCommand objects into
     * <code>commands</code>.
     */
    public RunControl (final String configurationFilename)
    {
        this (configurationFilename, NUM_OF_RETRIES_BEFORE_FAILURE_DEFAULT);
    }

    public RunControl (final String configurationFilename, final int maximumQueryResends)
    {
    	addCommand (new CompactPortalListCommand		 (this));
    	addCommand (new ClearPortalListCommand			 (this));
    	addCommand (new StartSocketsCommand              (this));
    	addCommand (new ListSocketsCommand               (this));
    	addCommand (new StopSocketsCommand               (this));
    	addCommand (new FlushSocketsCommand              (this));
        addCommand (new QuitCommand                      (this));        
        addCommand ("exit", new QuitCommand              (this)); // alias
        addCommand (new ListCommand                      (this));
        addCommand (new GetValueCommand                  (this));
        addCommand (new SetValueCommand                  (this));
        addCommand (new WaitUntilStableCommand           (this)); // DT specific
        addCommand (new PollStabilityCommand             (this)); // DT specific
        addCommand (new SetGridCoordsCommand             (this)); // DT specific
        addCommand (new SetNamesCommand                  (this)); // DT specific
        addCommand (new StartTranscriptCommand           (this));
        addCommand (new StopTranscriptCommand            (this));
        addCommand (new KillRemoteServersCommand         (this));
        //addCommand (new TimeoutValuesCommand             (this));

        addCommand (new HelpCommand                      (this));
        addCommand (new SleepCommand                     (this));
        addCommand (new DateCommand                      (this));
        addCommand (new MTimeCommand                     (this));
        addCommand (new EchoCommand                      (this));
        addCommand (new PauseCommand                     (this));
        addCommand (new SaveReadSchemaCommand            (this));
        addCommand (new SaveWriteSchemaCommand           (this));
        addCommand (new CheckClockSynchronizationCommand (this));
        addCommand (new SetPassiveModeCommand            (this));
        addCommand (new RunScriptCommand                 (this));
        addCommand (new GetExactStableTimeCommand        (this)); // DT specific
        addCommand (new GetExactStartTimeCommand         (this)); // DT specific
        
        config = HyperCastConfig.createConfig (configurationFilename);
        this.monitor = config.createMonitor();
        this.monitor.openMonitor();
        this.commandRunner = new CommandRunner (this.monitor, maximumQueryResends);

        config.log.println ("Monitor successfully created."); 
        
        socketsStatusTable = new SocketStatusTable(this);
    }
    
    public SocketStatusTable getSocketStatusTable(){
    	return this.socketsStatusTable;
    }
   

    /*
     * TODO: think about the design of this. This method added in the process of moving commands to individual files
     */
    public Hashtable getCommands()
    {
        return commands;
    }

    public I_Monitor getMonitor()
    {
        return monitor;
    }
    

    public CommandRunner getCommandRunner()
    {
        return commandRunner;
    }

    /* added by Guangyu to support Application-Experiment */
	/* 
	 * There're two types of experiment. 
	 * The RUNSERVER mode works only when RunServer is run
	 * Some commands are only designed for RunServer, so can only be executed in
	 * RUNSERVER mode
	 */
    public int experimentMode; // TODO: default?? private? accessor?
    public final static int MODE_RUNSERVER = 0;
    public final static int MODE_APPLICATION = 1;
    
    /**
     * This method is called to interrupt a pending command
     */
    public void interruptCommand()
    {
        commandRunner.interruptQuery();
    }
	
    /**
     * Used to insert an RunControlCommand object into <code>commands</code>.  
     * This may be modified in the future in order to achieve greater command functionality.  
     * For example, it could keep track of what order the commands were added in.
     *
     * TODO: augment this for reflection?
     */
    protected void addCommand (RunControlCommand commandImplementation)
    {
        commands.put (commandImplementation.name(), commandImplementation);
    }

    /**
     * @deprecated
     */
    protected void addCommand (String name, RunControlCommand commandImplementation)
    {
        commands.put (name, commandImplementation);
    }

    /**
     * Calls the <code>processCommand</code> method for each command. 
     * If a command throws an Exception("break"), then the processing halts.
     */
    public void processCommands (final String [] lines, final BufferedReader in, final PrintWriter out, final PrintWriter err)
        throws Exception
    { 
        try
        {
            for (int i = 0; i < lines.length; i++)
            {
                processCommand(lines[i], in, out, err);
            }
        }
        catch (Exception e)
        {
            if (!"break".equals(e.getMessage()))
                throw e;
        }
    }

    /**
     * @see processCommand (String,String,BufferedReader,PrintWriter,PrintWriter)
     */
    public boolean processCommand (String line, final BufferedReader in, PrintWriter out, final PrintWriter err)
		throws Exception
    {
        return processCommand ("UNRESTRICTED", line, in, out, err);
    }

    /**
     * Performs actions for a single command.
     * This should really be in the command line module, but is here because the gui is not full featured.
     */
    public boolean processCommand (final String protocolRestriction, String line, final BufferedReader in, PrintWriter out, final PrintWriter err)
		throws Exception
    {
		
		this.socketsStatusTable.err = err;
		this.socketsStatusTable.out = out;
    	
    	boolean outputRedirected = false;
		int redirect = line.indexOf ('>');
		if (redirect >= 0)
                {
			String redirectString = line.substring (redirect, line.length());
			line = line.substring(0, redirect);

			boolean append;
			String redirectFileName;
			if (redirectString.length() >= 2
                            &&
                            redirectString.charAt(1) == '>')
                        {
				append = true;
				redirectFileName = redirectString.substring (2, redirectString.length());
			}
                        else
                        {
				append = false;
				redirectFileName = redirectString.substring (1, redirectString.length());
			}

			int filenameStart = 0;
			while (filenameStart < redirectFileName.length()
				&&
                               Character.isWhitespace (redirectFileName.charAt (filenameStart)))
                        {
				filenameStart++;
                        }

			int filenameEnd = redirectFileName.length();
			while (filenameEnd - 1 > 0
				&&
                               Character.isWhitespace(redirectFileName.charAt(filenameEnd - 1)))
                        {
				filenameEnd--;
                        }

			if (filenameEnd == 0)
                        {
				err.println("No filename after '>' or '>>'!");
				out.println();
				return true;
			}

			try
                        {
				out =
                                    new PrintWriter(
                                        new FileWriter(
							redirectFileName.substring(filenameStart, filenameEnd),
							append));
				outputRedirected = true;
			}
                        catch (IOException ioe)
                        {
				err.println(
					"Exception occured while redirecting output to \""
						+ redirectFileName.substring(filenameStart, filenameEnd)
						+ "\" :"
						+ ioe.toString());
				out.println();
				return true;
			}
		}

		int pipe = line.indexOf('|');
		if (pipe >= 0)
                {
			err.println("Sorry, pipes are not supported.");
			out.println();
			return true;
		}

		StringTokenizer inputTokens = new StringTokenizer(line, " \t");
		String command = "";
		String firstArg = null, secondArg = null, thirdArg = null;

		if (inputTokens.hasMoreTokens())
			command = inputTokens.nextToken();

		int args = inputTokens.countTokens();
		if (inputTokens.hasMoreTokens()) {
			firstArg = inputTokens.nextToken();
		}
		if (inputTokens.hasMoreTokens()) {
			secondArg = inputTokens.nextToken();
		}
		if (inputTokens.hasMoreTokens()) {
			thirdArg = inputTokens.nextToken();
		}

		RunControlCommand commandObject = (RunControlCommand) commands.get(command);
		if (commandObject == null)
                {
                    throw new CommandFailureException ("There is no such command \"" + command + "\"");
		}
                else
                {
                    if (! protocolRestriction.equals ("UNRESTRICTED"))
                    {
                        if (! commandObject.validWithProtocol (protocolRestriction))
                        {
                            throw new CommandFailureException ("Command \"" + commandObject.name() + "\" cannot be run with protocol \"" + protocolRestriction + "\"");
                        }
                    }

			if (args + inputTokens.countTokens() < commandObject.minParameters()) {
				err.println(
					"This command requires a minimum of "
						+ commandObject.minParameters()
						+ " arguments.");
				err.println("USAGE: " + command + " " + commandObject.parameterDescription());
			} else if (commandObject.maxParameters() != RunControlCommand.UNLIMITED_ARG_COUNT
                                   &&
                                   args > commandObject.maxParameters()) {
				err.println(
					"This command accommodates a maximum of "
						+ commandObject.maxParameters()
						+ " arguments.");
				err.println("USAGE: " + command + " " + commandObject.parameterDescription());
			} else {
				try {
					commandObject.action(
						command,
						firstArg,
						secondArg,
						thirdArg,
						inputTokens,
						in,
						out,
						err);
				} catch (IllegalArgumentException iae) {
					err.println("Illegal argument in command usage: \"" + iae.toString() + "\"");
					err.println("USAGE: " + command + " " + commandObject.parameterDescription());
				}
                                catch (final InterruptedException ie)
                                {
                                    //err.println("RunControl.processCommand sees InterruptedException");
                                        throw ie;
				}

				if (outputRedirected) {
					out.flush();
					out.close();
				}
			}
		}
		out.println();
        return true;
	}
}
