/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/


package runcontrol;

import java.io.PrintWriter;
import java.io.BufferedReader;

import java.util.StringTokenizer;
import java.util.Random;
import java.util.Vector;

import java.lang.UnsupportedOperationException;

/** 
 * A parent class for all command objects in RunControl and derived classes.
 * 
 * @author HyperCast Team
 * @author Greg Mattes
 * @author Guangyu Dong
 * @version	2004-2005 
 * @author Michael Nahas
 * @version	2.0, Feb. 28, 2001
 */
public abstract class RunControlCommand 
{
    public static Random rand = new Random();

    /** Returns a simple description of the command's parameters, e.g. "(server|noserver) [port#]" */
    public abstract String parameterDescription();

    /** command string that a user of a command line interface types to invoke this command, e.g. CommandQuit is invoked with "quit" */
    public abstract String name();

    /** A one-line description of the command that can be printed by issuing the "help" command with no arguments. */
    public abstract String description();

    /** A multi-line detailed description of the command that can be printed by issuing the command "help <command>". */
    public abstract String longDescription();

    /** Returns the minimum number of arguments for this command. */
    public abstract int minParameters();
    
    /** Returns the maximum number of arguments for this command. */
    public static final int UNLIMITED_ARG_COUNT = -1;
    public abstract int maxParameters();

    /**
     * Returns false if this command cannot be run with the given
     * protocol. For example if a certain command can only be run with
     * DT then it should override this method and return false unless
     * restriction.equals ("DT").
     */
    public boolean validWithProtocol (final String restriction)
    {
        return true;
    }

    /**
     * Performs the action associated with this command.
     * @throws Exception
     *
     * @throws CommandFailureException if the command conforms to
     * modern exception handling and fails for any reason.
     *
     * @throws InterruptedException if the caller invokes
     * <code>interruptQuery</code> on the associated
     * <code>CommandRunner</code> object <i>and</i> the command is
     * written in such a way that it handles interruption properly,
     * i.e. it throws <code>InterruptedException</code> when
     * interrupted instead of suppressing the exception.
     *
     * @throws Exception This code was originally written to throw
     * objects of the general <code>Exception</code> class which is
     * not good practice.
     *
     * @deprecated use the new action method instead, note the change in the signature wrt exceptions
     */
    // TODO: throws Exception???
    
    public void action (String command,
                                 String firstArg,
                                 String secondArg,
                                 String thirdArg,
                                 StringTokenizer furtherArgs,
                                 BufferedReader in,
                                 PrintWriter out,
                                 PrintWriter err)
        throws Exception
    {
    	int argCount = 0;
    	
    	if(firstArg!=null)
    		argCount ++;
    	if(secondArg!=null)
    		argCount ++;
    	if(thirdArg!=null)
    		argCount ++;
    	
    	if(furtherArgs!=null)
    		argCount +=furtherArgs.countTokens();
    	
    	String []args = new String[argCount];
    	
    	if(args.length>0)
    		args[0] = firstArg;
    	if(args.length>1)
    		args[1] = secondArg;
    	if(args.length>2)
    		args[2] = thirdArg;
    	
    	int index = 3;
    	while(furtherArgs.hasMoreElements()){
    		args[index] = furtherArgs.nextToken();
    		index ++;
    	}
    	
    	this.action(command, args, in, out, err);
    }
    

	public void action(String command, String[] args, BufferedReader in, PrintWriter out, PrintWriter err)
            throws Exception
    { throw new UnsupportedOperationException ("This command has no implementation"); }

    // TODO: these should be static

    /** Converts an argument into an integer. */
    // TODO, allow a valid range to be specified
    protected int toInt (final String s) 
    {
        try 
        {
            return Integer.parseInt (s);
        }
        catch (NumberFormatException nfe)
        {
            throw new IllegalArgumentException ("Parameter not of type int.");
        }
    }

    /** Converts an argument into a long. */
    protected long toLong (final String s) 
    {
        try 
        {
            return Long.parseLong (s);
        }
        catch (NumberFormatException nfe)
        {
            throw new IllegalArgumentException ("Parameter not of type long.");
        }
    }

    /** Converts an argument into a floating-point value. */
    protected float toFloat (final String s) 
    {
        try 
        {
            return (Float.valueOf (s)).floatValue();
        }
        catch (NumberFormatException nfe)
        {
            throw new IllegalArgumentException("Parameter not of type float.");
        }
    }

    /**
     * Given the string: 1,5-9,10,3
     * Returns the array of ints: 1,5,6,7,8,9,10,3
     *
     * @throws IllegalArgumentException if the string cannot be parsed
     */
    protected int [] parseIndices (final String indicesString)
    {
        final Vector ints = new Vector();
        final StringTokenizer tokenizer = new StringTokenizer (indicesString, "-,", true);

        boolean inRange = false;
        int startRange = -1;
        int endRange = -1;

        boolean expectingNumber = true;

        String token = "";
        while (tokenizer.hasMoreTokens())
        {
            token = tokenizer.nextToken();
            //System.err.println (token);
            if (token.equals ("-"))
            {
                if (expectingNumber)
                    throw new IllegalArgumentException ("Parse error; was expecting number: " + token);

                expectingNumber = true;
                inRange = true;
            }
            else if (token.equals (","))
            {
                if (expectingNumber)
                    throw new IllegalArgumentException ("Parse error; was expecting number: " + token);

                if (inRange)
                {
                    if (startRange >= endRange)
                    {
                        throw new IllegalArgumentException ("Parse error; first range value must be less than second: " + token);
                    }

                    for (int i = startRange; i <= endRange; ++i)
                    {
                        Integer newInteger = new Integer (i);
                        if (! ints.contains (newInteger))
                        {
                            ints.add (newInteger);
                        }
                    }
                    endRange = -1;
                    inRange = false;
                }
                else
                {
                    Integer newInteger = new Integer (startRange);
                    if (! ints.contains (newInteger))
                    {
                        ints.add (newInteger);
                    }
                }
                expectingNumber = true;
            }
            else
            {
                if (! expectingNumber)
                    throw new IllegalArgumentException ("Parse error; was not expecting number: " + token);

                if (inRange)
                {
                    endRange = toInt (token);
                }
                else
                {
                    startRange = toInt (token);
                }
                expectingNumber = false;
            }
        }

        if (token.equals (",")
            ||
            token.equals ("-"))
        {
            throw new IllegalArgumentException ("Parse error; cannot end with: " + token);
        }

        if (endRange != -1)
        {
            if (startRange >= endRange)
            {
                throw new IllegalArgumentException ("Parse error; first range value must be less than second: " + token);
            }

            for (int i = startRange; i <= endRange; ++i)
            {
                Integer newInteger = new Integer (i);
                if (! ints.contains (newInteger))
                {
                    ints.add (newInteger);
                }
            }
        }
        else
        {
            Integer newInteger = new Integer (startRange);
            if (! ints.contains (newInteger))
            {
                ints.add (newInteger);
            }
        }

        int [] result = new int [ints.size()];
        for (int i = 0; i < ints.size(); ++i)
        {
            result [i] = ((Integer) ints.elementAt (i)).intValue();
        }

        return result;
    }
}
