/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/


package runcontrol;

import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Random;

import hypercast.*;
import hypercast.MonitorAndControl.I_Monitor;
import hypercast.MonitorAndControl.MonMessage;
import hypercast.MonitorAndControl.NoSuchPortalException;
import hypercast.util.XmlUtil;

import org.w3c.dom.Element;
/**
 * This is a data structure storing the running status of the ovelray sockets
 * created on remote RunServer.
 * 
 * User of RunControl can manupilate the sockets on RunServer to make them
 * either join or leave the network and the RunControl tries to locally keep
 * track the current status of every sockets in this table so that there is no
 * need to query the Portals every time the socket status information is needed.
 * 
 * It is possible that sometime this table is out of sync with the actual status
 * of the remote sockets due to the temporary or permanent loss of contact with
 * some RunServers. In this case, the corresponding status is marked
 * "Unsynchronized".
 * 
 * @author HyperCast Team
 * @author Guangyu Dong, January 2005
 *  
 */

public class SocketStatusTable {

	/** the list of the runserver socket status */
	private ArrayList socketStatusList;

	/** the runControl object. used to send command */
	private RunControl runControl;

	/** the monitor object used to update portal list and create query message */
	private I_Monitor monitor;
	
	/* package */ PrintWriter out, err;
	
	private long lastTimeStartingSockets = 0;
	
	private long lastTimeStoppingSockets = 0;
	
	/** the lock object */
	private Object lock = new Object();

	public SocketStatusTable(RunControl runControl){
		this.runControl = runControl;
		this.monitor = runControl.getMonitor();
		this.lastTimeStartingSockets = this.lastTimeStoppingSockets = System.currentTimeMillis();
		try {
			reset();
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	public Object getLock(){
		return lock;
	}
	
	/** to make this table compatible with the previous "Experiment" class 
	 * @throws InterruptedException*/
	public int [] getNodesOnServer() throws InterruptedException{
		updateSize();
		int [] nodesOnServer = new int [size()];
		
		for(int i=0;i<nodesOnServer.length;i++){
			nodesOnServer[i] = getRunningSocketCount(i);
		}
		
		return nodesOnServer;
	}
	
	class SlowStartPair{
		int portalIndex;
		int socketIndex;
		
		SlowStartPair(int pi, int si){
			portalIndex = pi;
			socketIndex = si;
		}
	}
	class StartMessageBatch{
		int []portals;
		MonMessage[] messages;
	}
	class SlowStartMessage{
		int portal;
		MonMessage message;
		int lastSocketIndex;
	}
	
	
	private int []getSlowStartSequence(int toStart[]) throws InterruptedException{
		ArrayList seq = new ArrayList();
		int leftToStart [] = new int [toStart.length];
		
		for(int i=0;i<toStart.length;i++){
			
			int running = this.getRunningSocketCount(i);
			leftToStart[i] = running;
			
			for(int j=0;j<toStart[i];j++){
				seq.add(new Integer(i));
			}
		}
		
		//randomize the sequence
		int new_seq[] = new int [seq.size()];
		int size = seq.size();
		
		Random r = new Random();
		for(int i=0;i<size;i++){
			int ran_index = r.nextInt(seq.size());
			new_seq[i] = ((Integer) seq.get(ran_index)).intValue();
			seq.remove(ran_index);
		}
		
		return new_seq;
	}
	private StartMessageBatch getStartMessages(int []toStart) throws InterruptedException{
		int msgCount = 0;
		for(int i=0;i<toStart.length;i++){
			if(toStart[i]>0)
				msgCount ++;
		}
		
		StartMessageBatch batch = new StartMessageBatch();
		batch.portals = new int[msgCount];
		batch.messages = new MonMessage[msgCount];
		
		int index=0;
		for(int i=0;i<toStart.length;i++){
			if(toStart[i]>0){
				batch.portals[index] = i;
				batch.messages[index] = monitor.createSetRequest();
				
				int running = this.getRunningSocketCount(i);
				for(int j=1;j<=toStart[i];j++){
					Element trueValue = XmlUtil.getXmlValue(batch.messages[index].getDocument(), "Running", "true");
					batch.messages[index].addStat(XmlUtil.createXPath("/Portal/Appl/Socket["+(j+running)+"]/Running"), trueValue);
				}
				index ++;
			}
		}
		
		return batch;
		
	}
	private StartMessageBatch getStopMessages(int []toSop) throws InterruptedException{
		int msgCount = 0;
		for(int i=0;i<toSop.length;i++){
			if(toSop[i]>0)
				msgCount ++;
		}
		
		StartMessageBatch batch = new StartMessageBatch();
		batch.portals = new int[msgCount];
		batch.messages = new MonMessage[msgCount];
		
		int index=0;
		for(int i=0;i<toSop.length;i++){
			if(toSop[i]>0){
				batch.portals[index] = i;
				batch.messages[index] = monitor.createSetRequest();
				
				int running = this.getRunningSocketCount(i);
				for(int j=0;j<toSop[i];j++){
					if((-j+running)>0){
						Element falseValue = XmlUtil.getXmlValue(batch.messages[index].getDocument(), "Running", "false");
						batch.messages[index].addStat(XmlUtil.createXPath("/Portal/Appl/Socket["+(-j+running)+"]/Running"), falseValue);
					}
				}
				index ++;
			}
		}
		
		return batch;
		
	}
	private StartMessageBatch getNextSlowStartMessages(int pairs[], int offset, int numToStart, boolean start) throws InterruptedException{
		ArrayList msgs = new ArrayList();
		int started = 0;
		while(offset<pairs.length && started<numToStart){
			
			boolean newMsg = true;
			//try to search a duplicate message on the same portal
			for(int i=0;i<msgs.size();i++){
				SlowStartMessage ssm = (SlowStartMessage)msgs.get(i);
				
				if(ssm.portal == pairs[offset]){
					newMsg = false;
					if(start){
						Element trueValue = XmlUtil.getXmlValue(ssm.message.getDocument(), "Running", "true");
						ssm.message.addStat(XmlUtil.createXPath("/Portal/Appl/Socket["+(ssm.lastSocketIndex+1)+"]/Running"), trueValue);
						ssm.lastSocketIndex++;
					}else{
						Element falseValue = XmlUtil.getXmlValue(ssm.message.getDocument(), "Running", "false");
						ssm.message.addStat(XmlUtil.createXPath("/Portal/Appl/Socket["+(ssm.lastSocketIndex-1)+"]/Running"), falseValue);
						ssm.lastSocketIndex--;
					}
					break;
				}
				
			}
			
			if(newMsg){
				SlowStartMessage ssm = new SlowStartMessage();
				ssm.portal = pairs[offset];
				ssm.message = monitor.createSetRequest();
				
				if(start){
					Element trueValue = XmlUtil.getXmlValue(ssm.message.getDocument(), "Running", "true");
					ssm.lastSocketIndex = this.getRunningSocketCount(ssm.portal)+1;
					ssm.message.addStat(XmlUtil.createXPath("/Portal/Appl/Socket["+ssm.lastSocketIndex+"]/Running"), trueValue);
					msgs.add(ssm);
				}else{
					Element falseValue = XmlUtil.getXmlValue(ssm.message.getDocument(), "Running", "false");
					ssm.lastSocketIndex = this.getRunningSocketCount(ssm.portal);
					ssm.message.addStat(XmlUtil.createXPath("/Portal/Appl/Socket["+ssm.lastSocketIndex+"]/Running"), falseValue);
					msgs.add(ssm);
				}
			}
			
			offset++;
			started ++;
		}
		
		StartMessageBatch batch = new StartMessageBatch();
		batch.portals = new int [msgs.size()];
		batch.messages = new MonMessage [msgs.size()];
		
		for(int i=0;i<msgs.size();i++){
			SlowStartMessage ssm = (SlowStartMessage)msgs.get(i);
			batch.portals[i] = ssm.portal;
			batch.messages[i] = ssm.message;
		}
		
		return batch;
		
	}
	
	/**
	 * start a number of sockets on available RunServers
	 * 
	 * @param count		number of sockets to start
	 * @param uniform   should the sockets be uniformly stated on different RunServers?
	 * @throws InterruptedException
	 * @throws IllegalArgumentException
	 */ 
	public int startSockets(int count, boolean uniform, int rate) throws IllegalArgumentException, InterruptedException{
 		// the next line resets the timer value, which may be read by wait_until_stable
                this.lastTimeStartingSockets = System.currentTimeMillis();
		synchronized(lock){
			if(uniform){
				return startNewSocketsUniform(count, rate);
			}else{
				return startNewSocketsInOrder(count, rate);
			}
		}
		
	}
	/**
	 * stop a number of sockets on RunServers
	 * 
	 * @param count		number of sockets to stop
	 * @param uniform   should the sockets be uniformly stopped on different RunServers?
	 * @throws InterruptedException
	 * @throws IllegalArgumentException
	 */ 
	public int stopSockets(int count, boolean uniform, int rate) throws IllegalArgumentException, InterruptedException{
		synchronized(lock){
			if(uniform){
				return stopSocketsUniform(count, rate);
				
			}else{
				return stopSocketsInOrder(count, rate);
			}
		}
		
	}
	
	private int startNewSockets(int toStartOnRS[]) throws InterruptedException{
		int started =0;
		StartMessageBatch batch = this.getStartMessages(toStartOnRS);
		MonMessage replies []=null;
		
		for(int i=0;i<toStartOnRS.length;i++){
			if(toStartOnRS[i]>0){
				SocketStatus status = get(i);
				status.setSync(false);
				status.setRunningSockets(status.getRunningSockets() + toStartOnRS[i]);
			}
		}
		
		try {
			replies = runControl.getCommandRunner().sendQueries(batch.portals, batch.messages);
		} catch (NoSuchPortalException e1) {
			err.println("Cannot start new sockets: "+e1.toString());
			return 0;
		}
		
		for(int i=0;i<replies.length;i++){
			SocketStatus status = get(batch.portals[i]);
			if(replies[i]==null){
				err.println("Message not acked by RS "+batch.portals[i]);
				status.setSync(false);
			}else{
				started+=toStartOnRS[batch.portals[i]];
				status.setSync(true);
			}
		}
		
		return started;
	}
	private int stopSockets(int toStopOnRS[]) throws InterruptedException{
		int stopped =0;
		StartMessageBatch batch = this.getStopMessages(toStopOnRS);
		MonMessage replies []=null;
		
		for(int i=0;i<toStopOnRS.length;i++){
			if(toStopOnRS[i]>0){
				SocketStatus status = get(i);
				status.setSync(false);
				status.setRunningSockets(status.getRunningSockets() - toStopOnRS[i]);
			}
		}
		
		try {
			replies = runControl.getCommandRunner().sendQueries(batch.portals, batch.messages);
		} catch (NoSuchPortalException e1) {
			err.println("Cannot stop new sockets: "+e1.toString());
			return 0;
		}
		
		for(int i=0;i<replies.length;i++){
			SocketStatus status = get(batch.portals[i]);
			if(replies[i]==null){
				err.println("Message not acked by RS "+batch.portals[i]);
				status.setSync(false);
			}else{
				stopped+=toStopOnRS[batch.portals[i]];
				status.setSync(true);
			}
		}
		
		return stopped;
	}
	
	private int startNewSocketsSlowly(int toStartOnRS[], int rate) throws InterruptedException{
		int started = 0;
		int [] portal_seq = this.getSlowStartSequence(toStartOnRS);
		int offset =0;
		
		while(offset<portal_seq.length){
			StartMessageBatch batch= this.getNextSlowStartMessages(portal_seq, offset, rate, true);
			
			for(int i=0;i<rate&&(offset+i)<portal_seq.length;i++){
				out.print(".");
				out.flush();
			}
			
			for(int i=0;i<batch.portals.length;i++){
				int portalIndex = batch.portals[i];
				SocketStatus status = get(portalIndex);
				status.setSync(false);
				status.setRunningSockets(status.getRunningSockets() + batch.messages[i].getStatsCount());
			}
			MonMessage replies[] = null;
			try {
				replies = runControl.getCommandRunner().sendQueries(batch.portals, batch.messages);
			} catch (NoSuchPortalException e1) {
				err.println("Cannot start new sockets: "+e1.toString());
				return 0;
			}
			
			for(int i=0;i<replies.length;i++){
				SocketStatus status = get(batch.portals[i]);
				if(replies[i]==null){
					err.println("Message not acked by RS "+batch.portals[i]);
					status.setSync(false);
				}else{
					started+=batch.messages[i].getStatsCount();
					status.setSync(true);
				}
			}
			
			offset += rate;
			
			if(offset< portal_seq.length -1){
				try{
					Thread.sleep(1000);
				}catch(InterruptedException e){
					e.printStackTrace();
				}
			}
			
		}
		
		out.println();
		return started;
	}
	
	private int stopSocketsSlowly(int toStopOnRS[], int rate) throws InterruptedException{
		int stopped = 0;
		int [] portal_seq = this.getSlowStartSequence(toStopOnRS);
		int offset =0;
		
		while(offset<portal_seq.length){
			StartMessageBatch batch= this.getNextSlowStartMessages(portal_seq, offset, rate, false);
			
			for(int i=0;i<rate&&(offset+i)<portal_seq.length;i++){
				out.print(".");
				out.flush();
			}
			
			for(int i=0;i<batch.portals.length;i++){
				int portalIndex = batch.portals[i];
				SocketStatus status = get(portalIndex);
				status.setSync(false);
				status.setRunningSockets(status.getRunningSockets() - batch.messages[i].getStatsCount());
			}
			MonMessage replies[] = null;
			try {
				replies = runControl.getCommandRunner().sendQueries(batch.portals, batch.messages);
			} catch (NoSuchPortalException e1) {
				err.println("Cannot stop new sockets: "+e1.toString());
				return 0;
			}
			
			for(int i=0;i<replies.length;i++){
				SocketStatus status = get(batch.portals[i]);
				if(replies[i]==null){
					err.println("Message not acked by RS "+batch.portals[i]);
					status.setSync(false);
				}else{
					stopped+=batch.messages[i].getStatsCount();
					status.setSync(true);
				}
			}
			
			offset += rate;
			
			if(offset< portal_seq.length -1){
				try{
					Thread.sleep(1000);
				}catch(InterruptedException e){
					e.printStackTrace();
				}
			}
			
		}
		
		out.println();
		return stopped;
	}
	
	/**
	 * 
	 * Uniformly start new sockets on RunServers
	 * 
	 * @param count       number of socket to start
	 * @return			actual number of socket started
	 * @return
	 * @throws IllegalArgumentException  if the number to start is larger than total avaialbe sockets.
	 * @throws InterruptedException
	 * @throws NoSuchPortalException
	 */
	private int startNewSocketsUniform(int count, int rate)throws IllegalArgumentException, InterruptedException{
		int available = getAvaiableSocketCount();
		if(available<count ){
			throw new IllegalArgumentException("Not enough available sockets to start. "+count+" is needed, "+available+" is available");
		}
		int toStart=count;
		int numOfRS = size();
		int toStartOnRS [] = new int[numOfRS];
		
		boolean startedInLastRound = true;
		
		for(int i=0;i<numOfRS; i++){
			toStartOnRS[i] = 0;
		}
		
		while (toStart>0 && startedInLastRound){
			startedInLastRound = false;
			
			for(int i=0;i<numOfRS;i++){
				try {
					if(monitor.isPortalActive(i)){
						SocketStatus status = get(i);
						int availableHere = status.getAvaialbleSockets() - toStartOnRS[i]; 
						if(availableHere>0){
							toStartOnRS[i] +=1;
							startedInLastRound = true;
							toStart --;
						}
					}
				} catch (NoSuchPortalException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
				if(toStart <=0)break;
			}
		}
		
		int started; 
		
		if(rate==0)
			started = startNewSockets(toStartOnRS);
		else
			started = startNewSocketsSlowly(toStartOnRS, rate);
		
		return started;
	}
	
	private int stopSocketsUniform(int count, int rate)throws IllegalArgumentException, InterruptedException{
		int running = this.getRunningSocketCount();
		if(running<count ){
			throw new IllegalArgumentException("Not enough sockets to stop. "+count+" is needed, "+running+" is running");
		}
		int toStop=count;
		int numOfRS = size();
		int toStopOnRS [] = new int[numOfRS];
		
		boolean stopInLastRound = true;
		
		for(int i=0;i<numOfRS; i++){
			toStopOnRS[i] = 0;
		}
		
		while (toStop>0 && stopInLastRound){
			stopInLastRound = false;
			
			for(int i=0;i<numOfRS;i++){
				try {
					if(monitor.isPortalActive(i)){
						SocketStatus status = get(i);
						int runningHere = status.getRunningSockets() - toStopOnRS[i]; 
						if(runningHere>0){
							toStopOnRS[i] +=1;
							stopInLastRound = true;
							toStop --;
						}
					}
				} catch (NoSuchPortalException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
				if(toStop <=0)break;
			}
		}
		
		int started; 
		
		if(rate==0)
			started = stopSockets(toStopOnRS);
		else
			started = stopSocketsSlowly(toStopOnRS, rate);
		
		return started;
	}
	
	/**
	 * Search through RunServers from low to high index to start a number of sockets
	 * 
	 * @param count		number of socket to start
	 * @return			actual number of socket started
	 * 
	 * @throws IllegalArgumentException		if the number to start is larger than total avaialbe sockets.
	 * @throws InterruptedException
	 */
	private int startNewSocketsInOrder(int count, int rate)throws IllegalArgumentException, InterruptedException{
		
		int available = getAvaiableSocketCount();
		if(available<count ){
			throw new IllegalArgumentException("Not enough available sockets to start. "+count+" is needed, "+available+" is available");
		}
		int numOfRS = size();
		int toStartOnRS[] = new int[numOfRS];
		for(int i=0;i<numOfRS; i++){
			toStartOnRS[i] = 0;
		}
		
		int toStart=count;
		for(int i=0;i<size();i++){
			
			try {
				if(monitor.isPortalActive(i)){
					SocketStatus status = get(i);
					int availableHere = status.getAvaialbleSockets(); 
					if(availableHere>0){
						int toStartHere = (toStart>availableHere) ? availableHere : toStart;
						//if(this.setRunningSocketCount(i, toStartHere + status.getRunningSockets()))
						toStartOnRS[i] = toStartHere;
						toStart -= toStartHere;
					}
						
				}
			} catch (NoSuchPortalException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			
			if(toStart<=0)break;
		}
		
		int started;
		
		if(rate==0)
			started = this.startNewSockets(toStartOnRS);
		else
			started = this.startNewSocketsSlowly(toStartOnRS, rate);
		
		return started;
	}
	
	/**
	 * search through RunServers from high to low index to stop some sockets.
	 * 
	 * @param count		number of sockets to stop
	 * @return			actual number of sockets that has been stopped
	 * 
	 * @throws IllegalArgumentException
	 * @throws InterruptedException
	 */
	private int stopSocketsInOrder(int count, int rate)throws IllegalArgumentException, InterruptedException{
		
		int running = getRunningSocketCount();
		if(running<count ){
			throw new IllegalArgumentException("Not enough running sockets to stop. "+count+" is needed, "+running+" is running");
		}
		
		int numOfRS = size();
		int toStopOnRS[] = new int[numOfRS];
		for(int i=0;i<numOfRS; i++){
			toStopOnRS[i] = 0;
		}
		
		int toStop=count;
		for(int i=size()-1;i>=0;i--){
			
			try {
				if(monitor.isPortalActive(i)){
					SocketStatus status = get(i);
					int runningHere = status.getRunningSockets(); 
					if(runningHere>0){
						int toStopHere = (toStop>runningHere) ? runningHere : toStop;
						toStopOnRS[i] = toStopHere;
						toStop -= toStopHere;
					}
						
				}
			} catch (NoSuchPortalException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			
			if(toStop<=0)break;
		}
		
		int stopped;
		
		if(rate==0)
			stopped = this.stopSockets(toStopOnRS);
		else
			stopped = this.stopSocketsSlowly(toStopOnRS, rate);
		
		return stopped;
		
		
	}
	
	
	
	/**
	 * get number of availalbe sockets to start.
	 * 
	 * sockets on dead RunServer is not counted.
	 * 
	 * @return
	 * @throws InterruptedException
	 */
	public int getAvaiableSocketCount() throws InterruptedException{
		
		return getTotalSocketCount() - getRunningSocketCount();
	}
	
	/**
	 * get running socket count on all RunServers.
	 * 
	 * sockets on dead RunServer is not counted.
	 * 
	 * @return
	 * @throws InterruptedException
	 */
	public int getRunningSocketCount() throws InterruptedException{
		updateSize();
		
		int running = 0;
		for(int i=0;i<size();i++){
			try{
				if(monitor.isPortalActive(i)){
					running += getRunningSocketCount(i);
				}
			}catch(NoSuchPortalException nspe){
				//should not happen after updateSize() is called
				break;
			}
		}
		
		return running;
	}
	
	/**
	 * Get total socket count created on all RunServers
	 * 
	 * sockets on dead RunServer is not counted.
	 * @throws InterruptedException
	 *
	 */
	public int getTotalSocketCount() throws InterruptedException{
		updateSize();
		
		int total = 0;
		for(int i=0;i<size();i++){
			try{
				if(monitor.isPortalActive(i)){
					total += getTotalSocketCount(i);
				}
			}catch(NoSuchPortalException nspe){
				//should not happen after updateSize() is called
				break;
			}
		}
		
		return total;
		
	}
	private MonMessage createStartMessage(int index, int runningCount){
		SocketStatus status = get(index);
		MonMessage query = monitor.createSetRequest();
		
		for(int i=1;i<=status.getTotalSockets();i++){
			Element value;
			if(i<runningCount){
				Element trueValue = XmlUtil.getXmlValue(query.getDocument(), "Running", "true");
				value = trueValue;
			}else{
				Element falseValue = XmlUtil.getXmlValue(query.getDocument(), "Running", "false");
				value = falseValue;
			}
			query.addStat(XmlUtil.createXPath("/Portal/Appl/Socket["+i+"]/Running"), value);
		}
		
		return query;
	}
	/** change the number of running sockets on the remote RunServer */
	public boolean setRunningSocketCount(int index, int runningCount) throws InterruptedException, NoSuchPortalException,
	NumberFormatException, HyperCastStatsException {
		CommandRunner cmdRunner = runControl.getCommandRunner();
		
		SocketStatus status = get(index);
		/** if the status is out of sync, sync it */
		if(!isSync(index)){
			sync(index);
		}
		
		if( runningCount > status.getTotalSockets() ){
			throw new IllegalArgumentException("Number of sockets to start is larger than the total socket number on RunServer "+ index);
		}
		
		if(runningCount==this.getRunningSocketCount(index))
			return true;
		
		if(runningCount>this.getRunningSocketCount(index)){
			this.lastTimeStartingSockets = System.currentTimeMillis();
		}else{
			this.lastTimeStoppingSockets = System.currentTimeMillis();
		}
		
		MonMessage query = this.createStartMessage(index, runningCount);
		
		status.setRunningSockets(runningCount);
		status.setSync(false);
		MonMessage reply = cmdRunner.sendQuery(index, query);
		if(reply!=null){
			status.setSync(true);
			return true;
		}else
			return false;
		
	}
	private SocketStatus get(int index){
		return (SocketStatus)this.socketStatusList.get(index);
	}
	
	/** start all sockets on RunServer 
	 * @throws InterruptedException
	 * @throws HyperCastStatsException
	 * @throws NumberFormatException*/
	public int startAllSockets() throws InterruptedException{
		contactAll();
		
		int started = 0;
		for(int i=0;i<size();i++){
			try {
				if(monitor.isPortalActive(i)){
					SocketStatus status = get(i);
					int availableHere = status.getAvaialbleSockets(); 
					if(availableHere>0){
						if(this.setRunningSocketCount(i, status.getTotalSockets()))
							started += availableHere;
					}
					
				}
			} catch (NumberFormatException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} catch (HyperCastStatsException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} catch (NoSuchPortalException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} 
			
		}
		
		return started;
	}
	
	/** stop all sockets on RunServer 
	 * @throws InterruptedException
	 * @throws */
	public int stopAllSockets() throws InterruptedException{
		contactAll();
		
		int stopped = 0;
		for(int i=0;i<size();i++){
			try {
				if(monitor.isPortalActive(i)){
					SocketStatus status = get(i);
					int runningHere = status.getRunningSockets(); 
					if(runningHere>0){
						if(this.setRunningSocketCount(i, 0))
							stopped += runningHere;
					}
					
				}
			} catch (NumberFormatException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} catch (HyperCastStatsException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} catch (NoSuchPortalException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} 
			
		}
		
		return stopped;
	}
	
	/** force all sockets on RunServer to stop, no matter what is the local status 
	 * @throws HyperCastStatsException
	 * @throws NumberFormatException
	 * @throws InterruptedException
	 * @throws NoSuchPortalException*/
	public void flushAllSockets() throws NumberFormatException, HyperCastStatsException, InterruptedException, NoSuchPortalException{
		
		synchronized(lock){
			updateSize();
			int []activePortals = monitor.getActivePortalIndices();
			MonMessage queries[] = new MonMessage[activePortals.length];
			
			//update number of sockets
			for(int i=0;i<queries.length;i++){
				queries [i]= monitor.createGetRequest();
				queries [i].addStat(XmlUtil.createXPath("/Portal/Appl/NumOfSockets"));
			}
			MonMessage replies [] = runControl.getCommandRunner().sendQueries(activePortals, queries);
				
			for(int i=0;i<replies.length;i++){
				if(replies[i]!=null ){
					SocketStatus status = get(activePortals[i]);
					if(!replies[i].isStatError(XmlUtil.createXPath("/Portal/Appl/NumOfSockets"))){
						int total = Integer.parseInt(XmlUtil.getScalarValue(replies[i].getStatValue(XmlUtil.createXPath("/Portal/Appl/NumOfSockets"))));
						status.setTotalSockets(total);
						status.setSync(true);
					}else{
						err.println("ERROR: statistics /Portal/Appl/NumOfSockets doesn't exist on RunServer "+activePortals[i]);
					}
				}else{
					err.println("ERROR: cannot update number of sockets on RunServer "+activePortals[i]);
				}
			}
						
			for(int i=0;i<queries.length;i++){
				SocketStatus status = get(activePortals[i]);
				queries [i]= this.createStartMessage(activePortals[i], 0);
				status.setSync(false);
				status.setRunningSockets(0);
			}
			
			replies = runControl.getCommandRunner().sendQueries(activePortals, queries);
				
			for(int i=0;i<replies.length;i++){
				if(replies[i]!=null){
					SocketStatus status = get(activePortals[i]);
					status.setSync(true);
				}else{
					err.println("ERROR: cannot flush sockets on RunServer "+activePortals[i]);
				}
			}
			
		}
		/*
		for(int i=0;i<size();i++){
			try {
				if(monitor.isPortalActive(i)){
					this.setRunningSocketCount(i, 0);
				}
			} catch (NumberFormatException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} catch (HyperCastStatsException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} catch (NoSuchPortalException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			
		}
		*/
	}
	
	/** portal list should be reset here to make the status table in sync 
	 * @throws InterruptedException*/
	public void clearPortalList() throws InterruptedException{
		monitor.clearPortalList();
		reset();
	}
	/** portal list should be compacted here to make the status table in sync */
	public void compactPortalList() throws NumberFormatException, HyperCastStatsException, InterruptedException {
		monitor.compactPortalList();
		reset();
		contactAll();
	}
	
	/**
	 * Get the total available socket count on the specified RunServer
	 * @param index
	 * @return
	 * @throws InterruptedException
	 */
	public int getAvailableSocketCount(int index) throws InterruptedException{
		updateSize();
		
		SocketStatus status = (SocketStatus) socketStatusList.get(index);
		
		return status.getAvaialbleSockets();
	}
	/**
	 * Get the total socket count on the specified RunServer
	 * @param index
	 * @return
	 * @throws InterruptedException
	 */
	public int getTotalSocketCount(int index) throws InterruptedException{
		updateSize();
		
		SocketStatus status = (SocketStatus) socketStatusList.get(index);
		
		return status.getTotalSockets();
	}
	/**
	 * Get the running socket count on the specified RunServer
	 * 
	 * @param index
	 * @return
	 * @throws InterruptedException
	 */
	public int getRunningSocketCount(int index) throws InterruptedException{
		updateSize();
		
		SocketStatus status = (SocketStatus) socketStatusList.get(index);
		
		return status.getRunningSockets();
	}
	/**
	 * check if the status of a specified RunServer is in sync
	 * 
	 * @param index
	 * @return
	 * @throws InterruptedException
	 */
	public boolean isSync(int index) throws InterruptedException{
		updateSize();
		
		SocketStatus status = (SocketStatus) socketStatusList.get(index);
		
		return status.isSync();
	}
	public boolean isActive(int index) throws NoSuchPortalException, InterruptedException{
		updateSize();
		return monitor.isPortalActive(index);
	}

	/**
	 * update the size of socket status list according to the current portal
	 * list.
	 * 
	 * This function should be called whenever the user try to access 
	 * the socket status, in order to make the size of status table
	 * equal to the size of the portal list;
	 * @throws InterruptedException
	 *  
	 */
	private void updateSize() throws InterruptedException {
		synchronized (lock){
			int[] portalindices = monitor.getAllPortalIndices();
			if (socketStatusList.size() > portalindices.length) {
				/* the portal list has shrinked. reset the socket status table */
				reset();
				contactAll();
			} else if (socketStatusList.size() < portalindices.length) {
				/* if new portal added, adds new socket status entry */
				for (int i = socketStatusList.size(); i < portalindices.length; i++) {
					socketStatusList.add(new SocketStatus(0, 0, false, false));
				}
				contactAll();
			}
		}
		
	}
	
	/**
	 * Synchronizes with all RunServers
	 * 
	 * @param force     determine if the already synchronized item need to be synchronized too
	 * @throws InterruptedException
	 * 
	 * @throws NoSuchPortalException
	 * @throws InterruptedException
	 * @throws HyperCastStatsException
	 * @throws NumberFormatException
	 */
	public void syncAll(boolean force) throws InterruptedException {
		synchronized (lock){
			updateSize();
			try{
				for(int i=0;i<size();i++){
					if(!isSync(i) || force)
						try {
							syncSingle(i);
						} catch (NumberFormatException e) {
							// TODO Auto-generated catch block
							e.printStackTrace();
						} catch (HyperCastStatsException e) {
							// TODO Auto-generated catch block
							e.printStackTrace();
						} 
				}
			}catch(NoSuchPortalException nspe){
				/* should not happend after updating the size */
			}
		}
	}
	private void contactAll() throws InterruptedException{
		synchronized (lock){
			updateSize();
			try{
				for(int i=0;i<size();i++){
					SocketStatus status = get(i);
					if(!status.isContacted())
						try {
							syncSingle(i);
						} catch (NumberFormatException e) {
							// TODO Auto-generated catch block
							e.printStackTrace();
						} catch (HyperCastStatsException e) {
							// TODO Auto-generated catch block
							e.printStackTrace();
						} 
				}
			}catch(NoSuchPortalException nspe){
				/* should not happend after updating the size */
			}
		}
	}
	public int size() throws InterruptedException{
		updateSize();
		return socketStatusList.size();
	}

	/**
	 * synchronizes this table with the actual status of sockets on a specific
	 * remote RunServer
	 * 
	 * @param index           the index of the RunServer to synchronize with
	 * 
	 * @throws NoSuchPortalException
	 * @throws InterruptedException
	 * @throws HyperCastStatsException
	 * @throws NumberFormatException
	 */
	public void sync(int poralIndex)throws InterruptedException, NoSuchPortalException,
			NumberFormatException, HyperCastStatsException {
		synchronized (lock){
			updateSize();
			syncSingle(poralIndex);
		}
		
	}
	private void syncSingle(int portalIndex) throws InterruptedException, NoSuchPortalException,
			NumberFormatException, HyperCastStatsException {
		synchronized (lock){
			SocketStatus status = (SocketStatus) socketStatusList.get(portalIndex);
			
			if(!monitor.isPortalActive(portalIndex)){
				status.setSync(false);
				return;
			}
			
			CommandRunner cmdRunner = runControl.getCommandRunner();
			MonMessage query = monitor.createGetRequest();
			query.addStat(XmlUtil.createXPath("/Portal/Appl/NumOfSockets"));
			//ask for status of each socket
			//TODO can be done in an easier way after I_Stats supports returning list
			for (int i = 1; i <= status.getTotalSockets(); i++) {
				query.addStat(XmlUtil.createXPath("/Portal/Appl/Socket[" + i + "]/Running"));
			}
			MonMessage reply;
			
			status.setSync(false);
			status.setContacted(true);
			reply = cmdRunner.sendQuery(portalIndex, query);
			
			if(reply == null) return;
			
			int total = Integer.parseInt(XmlUtil.getScalarValue(reply
					.getStatValue(XmlUtil.createXPath("/Portal/Appl/NumOfSockets"))));
			
			if(total != status.getTotalSockets()){
				//total is updated, need to sync again
				//TODO: can be change to use easier way after I_Stats is finished
				status.setTotalSockets(total);
				syncSingle(portalIndex);
				return;
			}
	
			int running = 0;
			for (int i = 1; i <= status.getTotalSockets(); i++) {
				if (!reply.isStatError(XmlUtil.createXPath("/Portal/Appl/Socket[" + i + "]/Running"))) {
					String rStr = XmlUtil.getScalarValue(reply
							.getStatValue(XmlUtil.createXPath("/Portal/Appl/Socket[" + i + "]/Running")));
					
					boolean r = Boolean.getBoolean(rStr);
					
					if(rStr.equals("true"))
						r = true;
					else
						r = false;
					
					if (r)
						running++;
				}
			}
	
			boolean sync = status.getTotalSockets() >= total ? true : running >= total;
	
			status.setTotalSockets(total);
			status.setRunningSockets(running);
			status.setSync(sync);
		}
	}

	/** reset the socket status table 
	 * @throws InterruptedException*/
	public void reset() throws InterruptedException {
		synchronized (lock){
			socketStatusList = new ArrayList();
			updateSize();
		}
	}

	/** this class store the sockets status of a single RunServer */
	class SocketStatus {
		/** the number of total sockets on this RunServer */
		private int TotalSockets;

		/** the number of running sockets on this RunServer */
		private int RunningSockets;

		/** if the status is in sync */
		private boolean sync;
		
		/** if this runserver has ever been contacted */
		private boolean contacted;

		public SocketStatus(int Total, int Running, boolean sync, boolean contacted) {
			this.TotalSockets = Total;
			this.RunningSockets = Running;
			this.sync = sync;
			this.contacted = contacted;
		}

		/**
		 * @return Returns the runningSockets.
		 */
		public int getRunningSockets() {
			return RunningSockets;
		}

		/**
		 * @param runningSockets
		 *            The runningSockets to set.
		 */
		public void setRunningSockets(int runningSockets) {
			RunningSockets = runningSockets;
		}

		/**
		 * @return Returns the sync.
		 */
		public boolean isSync() {
			return sync;
		}

		/**
		 * @param sync
		 *            The sync to set.
		 */
		public void setSync(boolean sync) {
			this.sync = sync;
		}

		/**
		 * @return Returns the totalSockets.
		 */
		public int getTotalSockets() {
			return TotalSockets;
		}
		
		/**
		 * @param totalSockets
		 *            The totalSockets to set.
		 */
		public void setTotalSockets(int totalSockets) {
			TotalSockets = totalSockets;
		}
		public int getAvaialbleSockets(){
			return getTotalSockets() - getRunningSockets();
		}

		/**
		 * @return Returns the contacted.
		 */
		public boolean isContacted() {
			return contacted;
		}
		/**
		 * @param contacted The contacted to set.
		 */
		public void setContacted(boolean contacted) {
			this.contacted = contacted;
		}
	}
	/**
	 * @return Returns the lastTimeStartingSockets.
	 */
	public long getLastTimeStartingSockets() {
		return lastTimeStartingSockets;
	}
	/**
	 * @return Returns the lastTimeStoppingSockets.
	 */
	public long getLastTimeStoppingSockets() {
		return lastTimeStoppingSockets;
	}
}
