/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

/* 
 * @author HyperCast Team
 */

// TODO Add comments 

package runcontrol.commands;

import java.io.BufferedReader;
import java.io.PrintWriter;

import java.util.StringTokenizer;

import hypercast.HyperCastConfig;
import hypercast.HyperCastStatsException;
import hypercast.MonitorAndControl.I_Monitor;
import hypercast.MonitorAndControl.MonMessage;
import hypercast.MonitorAndControl.NoSuchPortalException;
import hypercast.util.XmlUtil;

import hypercast.DT.DT_Node_ServerJoin;

import runcontrol.RunControl;
import runcontrol.RunControlCommand;
import runcontrol.CommandRunner;
import runcontrol.CommandFailureException;

/**
 * Polls RunServers participating in the current experiment to see
 * which sockets are stable. Non-blocking alternative to
 * <code>wait_until_stable</code> command.
 */
public final class PollStabilityCommand
    extends RunControlCommand
{
    private static final boolean debug = false;

    private MonMessage [] queries;
    private int [] portalIndices;

    private RunControl runControl;

    public PollStabilityCommand (RunControl runControl)
    {
        this.runControl = runControl;
    }

    public String parameterDescription() 
    {
        return "";
    }

    public String name()
    {
        return "poll_stability";
    }

    public boolean validWithProtocol (final String restriction)
    {
        return restriction.equals ("DT");
    }

    public String description() 
    {
        return "Displays the number of stable sockets on each run server for the current experiment.";
    }

    public String longDescription() 
    {
        return
            "Contacts each of the RunServers that are participating in the current experiment to see how many sockets have stabilized.";
    }

    public int minParameters()
    {
        return 0;
    }

    public int maxParameters()
    {
        return 0;
    }

    public void action (final String          command,
                        final String          firstArg,
                        final String          secondArg,
                        final String          thirdArg,
                        final StringTokenizer furtherArgs,
                        final BufferedReader  in,
                        final PrintWriter     out,
                        final PrintWriter     err)
        throws Exception
    {
        synchronized (runControl.getSocketStatusTable().getLock()) 
        {
            //final Experiment experiment = runControl.getExperiment();
        	int [] nodesOnServers = runControl.getSocketStatusTable().getNodesOnServer();
            makeMessages (nodesOnServers);
            final MonMessage [] replies = sendQueries();
            tabulateResults (replies, nodesOnServers, out, err);
        }     
    }
 
    private void tabulateResults (final MonMessage [] replies, int []nodesOnServers, final PrintWriter out, final PrintWriter err) throws HyperCastStatsException
    {
        out.println ("Experiment stability:");

        int totalStableCount = 0;
        int totalLeaderCount = 0;
        int totalSockets = 0;
        int totalErrors = 0;

        int j = 0;
        for (int i = 0; i < nodesOnServers.length; i++)
        {
            if (nodesOnServers [i] > 0)
            {
                int stableCountCurrentRunServer = 0;
                int leadersCountCurrentRunServer = 0;
                int errorsCurrentRunServer = 0;

                if (replies [j] == null)
                {
                    err.print ("Timeout from Portal index: " + i);
                }
                else
                {
                    for (int k = 0; k < replies[j].getStatsCount(); k++)
                    {
                        if (replies [j].isStatError (k))
                        {
                            err.print ("ERROR from Portal index: " + i + ": " + XmlUtil.getScalarValue(replies [j].getStatValue (k)));
                            ++errorsCurrentRunServer;
                        }
                        final String stateString = XmlUtil.getScalarValue(replies [j].getStatValue (k));
                        if (stateString.length() > 0) // TODO: include check that statistic is actually the state stat
                        {
                            try
                            {
                                final int state = Integer.parseInt (stateString);
                            
                                switch (state)
                                {
                                    case DT_Node_ServerJoin.LEADER_STABLE:
                                        ++leadersCountCurrentRunServer;
                                        // fall-through: leader stable counts as stable
                                    
                                    case DT_Node_ServerJoin.STABLE:
                                        ++stableCountCurrentRunServer;
                                        break;
                                    
                                    default:
                                        // do nothing: not stable
                                }
                            }
                            catch (final NumberFormatException nfe)
                            {
                                err.print ("ERROR from Portal index: " + i + ", state of socket " + k + " is not integer: " + stateString);
                                ++errorsCurrentRunServer;
                            }
                        }
                    }
                }

                totalLeaderCount += leadersCountCurrentRunServer;
                totalStableCount += stableCountCurrentRunServer;
                int socketsCurrentRunServer = nodesOnServers [i];
                totalSockets += socketsCurrentRunServer;
                totalErrors += errorsCurrentRunServer;

                final String leaderString = (leadersCountCurrentRunServer > 0) ? " (" + leadersCountCurrentRunServer + " leaders)" : "";
                final String errorString = (errorsCurrentRunServer > 0) ? " (" + errorsCurrentRunServer + " errors)" : "";
                out.println ("Portal " + i + ": " + stableCountCurrentRunServer + "/" + socketsCurrentRunServer + leaderString + errorString);

                ++j;
            }
        }

        final String leaderString = (totalLeaderCount > 1) ? " (" + totalLeaderCount + " leaders)" : "";
        final String errorString = (totalErrors > 0) ? " (" + totalErrors + " errors)" : "";
        final float stablePercentage = (((float) totalStableCount) / ((float) totalSockets)) * 100F;
        out.println ("Total: " + totalStableCount + "/" + totalSockets + "; " + stablePercentage + "% stable " + leaderString + errorString);
    }

    private MonMessage [] sendQueries()
        throws Exception
    {
        final MonMessage [] replies;
        final CommandRunner commandRunner = runControl.getCommandRunner();
        try
        {
            replies = commandRunner.sendQueries (portalIndices, queries); // blocks
        }
        catch (final NoSuchPortalException nspe)
        {
            nspe.printStackTrace();
            throw new CommandFailureException ("NoSuchPortalException caught: ", nspe);
        }
        return replies;
    }


    // Todo: this should stored with the experiment object and available in O(1) time - need to make all members of Experiment encapsulated before this could be trusted.
    private int getParticipatingRunServerCount (final int []nodesOnServers)
    {
        int participatingRunServerCount = 0;
        for (int i = 0; i < nodesOnServers.length; ++i)
        {
            if (nodesOnServers [i] > 0)
            {
                ++participatingRunServerCount;
            }
        }
        return participatingRunServerCount;
    }

    private void makeMessages (int []nodesOnServers)
    {
        final int participatingRunServerCount = getParticipatingRunServerCount (nodesOnServers);

        final I_Monitor monitor = this.runControl.getMonitor();
        queries = new MonMessage [participatingRunServerCount];
        portalIndices = new int [participatingRunServerCount];

        int j = 0;
        for (int i = 0; i < nodesOnServers.length; ++i)
        {
            if (nodesOnServers[i] > 0)
            {
                portalIndices [j] = i;
                queries [j] = monitor.createGetRequest();
                for (int k = 1; k <= nodesOnServers [i]; k++)
                {
                    queries [j].addStat (XmlUtil.createXPath("/Portal/Appl/Socket[" + k + "]/Node/State"));
                }
                ++j;
            }
        }
    }
}
