/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

/* 
 * @author HyperCast Team
 */

// TODO Add comments 

package runcontrol.commands;

import java.io.BufferedReader;
import java.io.PrintWriter;

import java.util.Random;
import java.util.StringTokenizer;

import hypercast.MonitorAndControl.I_Monitor;
import hypercast.MonitorAndControl.MonMessage;
import hypercast.util.XmlUtil;
import hypercast.util.XmlUtil;

import hypercast.HyperCastConfig;
import hypercast.HyperCastStatsException;

import runcontrol.RunControl;
import runcontrol.RunControlCommand;
import runcontrol.CommandRunner;
import runcontrol.CommandFailureException;

import gui.XYAddress;

public final class SetGridCoordsCommand
    extends RunControlCommand
{
    private static final boolean debug = false;

    private RunControl runControl;

    private final static int RANDOM_SHIFT_FACTOR = 20;

    public SetGridCoordsCommand (final RunControl runControl) 
    {
        this.runControl = runControl;
    }

    public String description() 
    {
        return "assigns DT logical address so that nodes form a square grid.";
    }

    public String longDescription() 
    {
        return "sets the coordinates of a DT network so that the nodes are arranged into a square grid structure.";
    }

    public String parameterDescription()
    {
        return "A single string of the form: 5x5x100x100x100x100 - NumberOfGridRows x NumberOfGridColumns x StartXCoordinate x StartYCoordinate x NeighborColumnDistance x NeighborRowDistance\n" +
            "followed optionally by a random shift factor (default 20), followed optionally by the string \"uniform\" which means that successive entries of the grid should use a socket on the next runserver if multiple runservers are available";
    }

    public int minParameters()
    {
        return 1;
    }

    public int maxParameters()
    {
        return 3;
    }

    public boolean validWithProtocol (final String restriction)
    {
        return restriction.equals ("DT");
    }

    public String name()
    {
        return "set_grid_coords";
    }

    public void action (final String          command,
                        final String          firstArg,
                        final String          secondArg,
                        final String          thirdArg,
                        final StringTokenizer furtherArgs,
                        final BufferedReader  in,
                        final PrintWriter     out,
                        final PrintWriter     err)
        throws Exception
    {
        synchronized (runControl.getSocketStatusTable().getLock()) 
        {
           
            final int [] nodesOnServers = runControl.getSocketStatusTable().getNodesOnServer();

            /*
             * Gather information on RunServers that will be used to build a grid.
             */

            // This tells us how many messages we need to send.
            int allocatedRunServerCount = countAllocatedRunServers (nodesOnServers);

            // This tells us which RunServers to use.
            int [] allocatedRunServerIndices = getAllocatedRunServerIndices (nodesOnServers, allocatedRunServerCount);

            // placing on a grid exactly will lead to 4 sockets on a
            // circle, which breaks DT. This random shift lowers the
            // chance of 4 sockets on a circle.
            int randomShift = (secondArg != null) ? toInt (secondArg) : RANDOM_SHIFT_FACTOR;

            // Uniform means that the first grid coord is assigned to
            // a socket on the first runserver, the second grid coord
            // is assigned to a socket on the second runserver,
            // etc. Not uniform means use all sockets on the first run
            // server, then use all sockets on the next runserver,
            // etc.
            boolean allocUniform = (thirdArg != null) && thirdArg.equalsIgnoreCase ("uniform");

            /*
             * Build messages for RunServers that set the logical addresses of nodes for a grid layout.
             */

            /*
             * Each node that participates in the two dimensional grid
             * is viewed a single dimension (linear) array of
             * nodes. As we build the grid we keep track of how far we
             * have progressed through this list of nodes.
             *
             * For example if we are making a 3x4 grid using 2
             * RunServers and each RunServer has 6 nodes, this loop
             * would execute twice, the first iteration would allocate
             * the first 6 nodes with startLinearPosition set to 0,
             * the second iteration would allocate the other 6 nodes
             * with startLinearPosition set to 6:
             *
             * 1 1 1 1
             * 1 1 6 6
             * 6 6 6 6
             *
             * The actual linear positions of the nodes are:
             *
             * 0  1  2  3
             * 4  5  6  7
             * 8  9 10 11
             */
            int startLinearPosition = 0;

            final MonMessage [] setLogicalAddressMessages = new MonMessage [allocatedRunServerCount];
            final I_Monitor monitor = runControl.getMonitor();

            if (allocUniform)
            {
                if (debug)
                    out.println ("doing uniform");

                if (allocatedRunServerCount == 0)
                        throw new CommandFailureException ("not enough sockets! in fact, no sockets at all!");

                final int [] remainingSocketCount = new int [allocatedRunServerCount];
                for (int i = 0; i < allocatedRunServerCount; ++i)
                    remainingSocketCount [i] = nodesOnServers [allocatedRunServerIndices [i]];

                StringTokenizer st = new StringTokenizer (firstArg, "x");

                final int rowCount    = toInt (st.nextToken().trim());
                final int columnCount = toInt (st.nextToken().trim());

                final int gridCount = rowCount * columnCount;

                if (debug)
                    out.println ("gridCount: " + gridCount);

                int runserverIndex = 0;

                for (; startLinearPosition < gridCount; ++startLinearPosition)
                {
                    final String coordInput = firstArg + (firstArg.endsWith ("x") ? "" : "x") + startLinearPosition + "x" + 1;
                    final XYAddress [] coordinateArray = getCoordinates (coordInput, randomShift);

                    if (debug)
                        out.println ("runserverIndex: " + runserverIndex + " remainingSocketCount [runserverIndex]: " + remainingSocketCount [runserverIndex]);

                    if (--remainingSocketCount [runserverIndex] < 0)
                        throw new CommandFailureException ("not enough sockets!");

                    if (setLogicalAddressMessages [runserverIndex] == null)
                        setLogicalAddressMessages [runserverIndex] = monitor.createSetRequest();

                    final int xCoord = coordinateArray [0].getX();
                    final int yCoord = coordinateArray [0].getY();
                    final String logicalAddressString = xCoord + "," + yCoord;

                    if (debug)
                        out.println ("logicalAddressString: " + logicalAddressString);

                    final int socketIndex = nodesOnServers[ allocatedRunServerIndices [runserverIndex]] - remainingSocketCount [runserverIndex];
                    final String logicalAddressXPath = "/Portal/Appl/Socket[" + socketIndex + "]/Node/LogicalAddress";

                    if (debug)
                        out.println ("socketIndex: " + socketIndex);

                    setLogicalAddressMessages [runserverIndex].addStat (XmlUtil.createXPath (logicalAddressXPath),
                                                                        XmlUtil.getXmlValue (XmlUtil.createDocument(), logicalAddressXPath, logicalAddressString));

                    runserverIndex = (runserverIndex + 1) % allocatedRunServerIndices.length;
                }

                if (debug)
                    throw new CommandFailureException ("just testing");
            }
            else
            {
                if (debug)
                    out.println ("not doing uniform");

                for (int i = 0; i < allocatedRunServerIndices.length; startLinearPosition += nodesOnServers[ allocatedRunServerIndices [i]], i++)
                {
                    if (nodesOnServers [ allocatedRunServerIndices [i]] <= 0)
                        throw new IllegalArgumentException ("RunServer with no nodes allocated to this experiment encountered when calculating grid coordinates");

                    final String coordInput = firstArg + (firstArg.endsWith ("x") ? "" : "x") + startLinearPosition + "x" + nodesOnServers[ allocatedRunServerIndices [i]];
                    final XYAddress [] coordinateArray = getCoordinates (coordInput, randomShift);
 							
                    setLogicalAddressMessages[i] = monitor.createSetRequest();
                    for (int k = 0; k < nodesOnServers [allocatedRunServerIndices [i]]; k++) 
                    {
                        final int xCoord = coordinateArray[k].getX();
                        final int yCoord = coordinateArray[k].getY();
                        final String logicalAddressString = xCoord + "," + yCoord;
                        final String logicalAddressXPath = "/Portal/Appl/Socket[" + (k+1) + "]/Node/LogicalAddress";
                        setLogicalAddressMessages[i].addStat(XmlUtil.createXPath(logicalAddressXPath), XmlUtil.getXmlValue(XmlUtil.createDocument(), logicalAddressXPath, logicalAddressString));
                        //RunControl.addScalarStatsIntoMsg(setLogicalAddressMessages[i], logicalAddressXPath, logicalAddressString);
                    }
                }
            }

            /*
             * All messages are prepared, send them to all RunServers and wait for response.
             */
            final CommandRunner commandRunner = runControl.getCommandRunner();
            final MonMessage [] result = commandRunner.sendQueries (allocatedRunServerIndices, setLogicalAddressMessages); // blocks on receipt or timeout of messages

            /*
             * Scan results for timeouts and errors.
             */
            for (int i = 0; i < allocatedRunServerIndices.length; i++)
            {
                if (result[i] == null)
                {
                    err.println ("Message was not acknowledged (timeout occurred) by RunServer with index " + allocatedRunServerIndices[i]);
                }
                else
                {
                    for (int k = 0; k < result[i].getStatsCount(); ++k)
                    {
                        if (result[i].isStatError (k))
                        {
                            err.println ("ERROR at runserver "
                                         + allocatedRunServerIndices[i]
                                         + ": "
                                         + XmlUtil.getScalarValue(result[i].getStatValue (k)));
                        }
                    }
                }
            }
        }
    }

    /**
     * Counts how many run servers are used? *not* the number of
     * sockets on run servers.
     */
    private int countAllocatedRunServers (final int [] nodesOnServers)
    {
        int allocatedRunServerCount = 0;
        for (int i = 0; i < nodesOnServers.length; ++i)
        {
            if (nodesOnServers[i] > 0)
            {
                ++allocatedRunServerCount;
            }
        }
        return allocatedRunServerCount;
    }

    /**
     * @see #getAllocatedRunServerIndices(int[],int)
     */
    private int [] getAllocatedRunServerIndices (final int [] nodesOnServers)
    {
        return getAllocatedRunServerIndices (nodesOnServers, -1);
    }

    /**
     * Generates an array that contains the indices of the RunServers
     * that are allocated to this experiment. There are a number of
     * RunServers that the RunControl knows about. Some subset are
     * allocated to this experiment.
     *
     * @param allocatedRunServerCount If the caller knows how many
     * RunServers are allocated to this experiment it can pass this
     * value as a hint to this method.
     */
    private int [] getAllocatedRunServerIndices (final int [] nodesOnServers, int allocatedRunServerCount)
    {
        if (allocatedRunServerCount <= 0)
        {
            allocatedRunServerCount = countAllocatedRunServers (nodesOnServers);
        }

        final int [] allocatedRunServerIndices = new int [allocatedRunServerCount];
        for (int i = 0, j = 0; i < nodesOnServers.length; ++i)
        {
            if (nodesOnServers[i] > 0)
            {
                allocatedRunServerIndices[j] = i;
                ++j;
            }
        }
        return allocatedRunServerIndices;
    }

    /**
     * @param input The input string has the following format:
     * [Grid Row Count]x[Grid Column Count]x[Start X Coordinate]x[Start Y Coordinate]x
     * [X Direction internode distance]x[Y direction internode distance]x[Linear Start Position]x[number of nodes on this RunServer]
     *
     * For example 10x15x500x500x100x100x4x15
     */
    private XYAddress [] getCoordinates (final String input, final int randomShift)
    {
        StringTokenizer st = new StringTokenizer (input, "x");

        if (st.countTokens() != 8)
            throw new IllegalArgumentException ("The input to set the grid coordinate is incorrect: " + input);

        final int rowCount    = toInt (st.nextToken().trim());
        final int columnCount = toInt (st.nextToken().trim());
 		
        final int columnStart      = toInt (st.nextToken().trim());
        final int rowStart         = toInt (st.nextToken().trim());
        final int interColumnSpace = toInt (st.nextToken().trim());
        final int interRowSpace    = toInt (st.nextToken().trim());
 		
        int oneDimensionPosition = toInt (st.nextToken().trim());

        final int nodeCount = toInt (st.nextToken().trim());

        final XYAddress [] addressArray = new XYAddress [nodeCount];
 		
        for (int i = 0; i < nodeCount; ++i, ++oneDimensionPosition)
        {
            final int row = oneDimensionPosition / columnCount;
            final int column = oneDimensionPosition % columnCount;
 			
            final int nextLogicalXCoord =
                columnStart +
                column * interColumnSpace +
                ((int) (randomShift * rand.nextDouble()));
//               + ((row % 2 == 0) ? 0 : interColumnSpace / 2); // add this calculation to create triangles instead of squares

            final int nextLogicalYCoord =
                rowStart +
                row * interRowSpace +
                ((int) (randomShift * rand.nextDouble()));

            addressArray[i] = new XYAddress (nextLogicalXCoord, nextLogicalYCoord);
        }
        return addressArray;
    }
}
