/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

/* 
 * @author HyperCast Team
 */

// TODO Add comments 

package runcontrol.commands;

import java.io.BufferedReader;
import java.io.PrintWriter;

import java.util.Random;
import java.util.StringTokenizer;

import hypercast.MonitorAndControl.I_Monitor;
import hypercast.MonitorAndControl.MonMessage;
import hypercast.util.XmlUtil;
import hypercast.util.XmlUtil;

import hypercast.HyperCastConfig;
import hypercast.HyperCastStatsException;

import runcontrol.RunControl;
import runcontrol.RunControlCommand;
import runcontrol.CommandRunner;
import runcontrol.CommandFailureException;

import gui.XYAddress;

public final class SetNamesCommand
    extends RunControlCommand
{
    private RunControl runControl;

    public SetNamesCommand (final RunControl runControl) 
    {
        this.runControl = runControl;
    }

    public String description() 
    {
        return "assigns a random name to all sockets.";
    }

    public String longDescription() 
    {
        return "randomly sets the names of all sockets.";
    }

    public String parameterDescription()
    {
        return "set_names [<portal index> <socket index> <name>]";
    }

    public int minParameters()
    {
        return 0;
    }

    public int maxParameters()
    {
        return 3;
    }

    public boolean validWithProtocol (final String restriction)
    {
        return restriction.equals ("DT");
    }

    public String name()
    {
        return "set_names";
    }

    public void action (final String          command,
                        final String          firstArg,
                        final String          secondArg,
                        final String          thirdArg,
                        final StringTokenizer furtherArgs,
                        final BufferedReader  in,
                        final PrintWriter     out,
                        final PrintWriter     err)
        throws Exception
    {
        synchronized (runControl.getSocketStatusTable().getLock()) 
        {
            final MonMessage [] setNameMessages;
            final int [] allocatedRunServerIndices;

            final I_Monitor monitor = runControl.getMonitor();

            final String nameXPath = "/Portal/Appl/SetName";

            if (firstArg != null
                &&
                secondArg != null
                &&
                thirdArg != null)
            {
                setNameMessages = new MonMessage [] { monitor.createSetRequest() };
                allocatedRunServerIndices = new int [] { toInt (firstArg) };
                setNameMessages[0].addStat (XmlUtil.createXPath (nameXPath), XmlUtil.getXmlValue(XmlUtil.createDocument(), nameXPath, secondArg + ":" + thirdArg));
            }
            else
            {
                final int [] nodesOnServers = runControl.getSocketStatusTable().getNodesOnServer();
                final int [] randomArray = generateRandomArray (countSockets (nodesOnServers));

                // This tells us how many messages we need to send.
                int allocatedRunServerCount = countAllocatedRunServers (nodesOnServers);

                // This tells us which RunServers to use.
                allocatedRunServerIndices = getAllocatedRunServerIndices (nodesOnServers, allocatedRunServerCount);

                /*
                 * Build messages for RunServers that set the names of sockets
                 */
                setNameMessages = new MonMessage [allocatedRunServerCount];

                int socketCounter = 0;
                for (int i = 0; i < allocatedRunServerIndices.length; i++)
                {
                    if (nodesOnServers [allocatedRunServerIndices [i]] <= 0)
                        throw new IllegalArgumentException ("RunServer with no nodes allocated to it when setting names");

                    setNameMessages[i] = monitor.createSetRequest();

                    final StringBuffer nameStringbuffer = new StringBuffer();

                    for (int k = 0; k < nodesOnServers [allocatedRunServerIndices [i]]; k++) 
                    {
                        nameStringbuffer.append (zeroPaddedString (randomArray [socketCounter], 4) + ",");
                        ++socketCounter;
                    }

                    setNameMessages[i].addStat (XmlUtil.createXPath (nameXPath), XmlUtil.getXmlValue(XmlUtil.createDocument(), nameXPath, nameStringbuffer.toString()));
                }
            }
            /*
             * All messages are prepared, send them to all RunServers and wait for response.
             */
            final CommandRunner commandRunner = runControl.getCommandRunner();

            out.println ("Sending set name messages to " + allocatedRunServerIndices.length + " servers");
            MonMessage [] result = commandRunner.sendQueries (allocatedRunServerIndices, setNameMessages); // blocks on receipt or timeout of messages

            /*
             * Scan results for timeouts and errors.
             */
            for (int i = 0; i < allocatedRunServerIndices.length; i++)
            {
                if (result[i] == null)
                {
                    err.println ("Message was not acknowledged (timeout occurred) by RunServer with index " + allocatedRunServerIndices[i]);
                }
                else
                {
                    for (int k = 0; k < result[i].getStatsCount(); ++k)
                    {
                        if (result[i].isStatError (k))
                        {
                            err.println ("ERROR at runserver "
                                         + allocatedRunServerIndices[i]
                                         + ": "
                                         + XmlUtil.getScalarValue(result[i].getStatValue (k)));
                        }
                    }
                }
            }
        }
    }

    /**
     * Returns an array of the specified size with each element being
     * an integer in the range [0,size-1] randomly permuted.
     */
    private int [] generateRandomArray (final int size)
    {
        final java.util.Random random = new java.util.Random();
        final int [] array = new int [size];

        for (int i = 0; i < size; ++i) 
            array[i] = i;

        for (int i = 0; i < size - 1; ++i)
        {
            final int randomIndex = i + random.nextInt (size - i);
            final int temp = array [i];
            array [i] = array [randomIndex];
            array [randomIndex] = temp;
        }

        return array;
        //for (int i = 0; i < size; ++i) System.out.print (array[i] + " ");
        //System.out.print ("\n");
    }

    /**
     * Counts how many run servers are used? *not* the number of
     * sockets on run servers.
     */
    private int countAllocatedRunServers (final int [] nodesOnServers)
    {
        int allocatedRunServerCount = 0;
        for (int i = 0; i < nodesOnServers.length; ++i)
        {
            if (nodesOnServers[i] > 0)
            {
                ++allocatedRunServerCount;
            }
        }
        return allocatedRunServerCount;
    }

    /**
     * Returns the total number of sockets on all runservers.
     */
    private int countSockets (final int [] nodesOnServers)
    {
        int socketCount = 0;
        for (int i = 0; i < nodesOnServers.length; ++i)
            if (nodesOnServers[i] > 0)
                socketCount += nodesOnServers[i];

        return socketCount;
    }

    /**
     * @see #getAllocatedRunServerIndices(int[],int)
     */
    private int [] getAllocatedRunServerIndices (final int [] nodesOnServers)
    {
        return getAllocatedRunServerIndices (nodesOnServers, -1);
    }

    /**
     * Generates an array that contains the indices of the RunServers
     * that are allocated to this experiment. There are a number of
     * RunServers that the RunControl knows about. Some subset are
     * allocated to this experiment.
     *
     * @param allocatedRunServerCount If the caller knows how many
     * RunServers are allocated to this experiment it can pass this
     * value as a hint to this method.
     */
    private int [] getAllocatedRunServerIndices (final int [] nodesOnServers, int allocatedRunServerCount)
    {
        if (allocatedRunServerCount <= 0)
        {
            allocatedRunServerCount = countAllocatedRunServers (nodesOnServers);
        }

        final int [] allocatedRunServerIndices = new int [allocatedRunServerCount];
        for (int i = 0, j = 0; i < nodesOnServers.length; ++i)
        {
            if (nodesOnServers[i] > 0)
            {
                allocatedRunServerIndices[j] = i;
                ++j;
            }
        }
        return allocatedRunServerIndices;
    }

    /**
     * Given an integer (int) and a length returns a string of the
     * given length containing the integer with prefix padding of
     * 0's. For example: zeroPaddedString (5, 3) -> "005"
     */
    private String zeroPaddedString (final int anInt, final int length)
    {
        final String orig = Integer.toString (anInt);
        final int numOfLeadingZeros = length - orig.length();
        if (numOfLeadingZeros < 0) { throw new RuntimeException(); }
        final StringBuffer stringBuffer = new StringBuffer (length);
        for (int j = 0; j < length; ++j) { stringBuffer.append ('0'); }
        for (int j = numOfLeadingZeros; j < length; ++j) { stringBuffer.setCharAt (j, orig.charAt (j - numOfLeadingZeros)); }
        return stringBuffer.toString();
    }
}
