/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

/* 
 * @author HyperCast Team
 */

// TODO Add comments 

package runcontrol.commands;

import java.io.BufferedReader;
import java.io.PrintWriter;

import java.util.StringTokenizer;

import hypercast.HyperCastConfig;
import hypercast.HyperCastStatsException;
import hypercast.MonitorAndControl.I_Monitor;
import hypercast.MonitorAndControl.MonMessage;
import hypercast.MonitorAndControl.NoSuchPortalException;
import hypercast.util.XmlUtil;

import hypercast.DT.DT_Node_ServerJoin;

import runcontrol.RunControl;
import runcontrol.RunControlCommand;
import runcontrol.CommandRunner;
import runcontrol.CommandFailureException;

// shouldn't this command be written like this: I (the runcontrol)
// am interested in knowing when the sockets of my experiment are
// stable, so I (the runcontrol) will ask all of you (runservers)
// to tell me when the sockets that you manage for me are stable.
//
// Doing things that way would reduce a lot of network traffic...

public final class WaitUntilStableCommand
    extends RunControlCommand
{
    private static final boolean debug = false;

    /** The time between checks that the Delaunay is stable. In seconds.  */
    public static final int DEFAULT_STABLE_RESOLUTION = 2; // seconds

    private static final float DEFAULT_STABILITY_FACTOR = 0.95f;

    private RunControl runControl;

    public WaitUntilStableCommand (RunControl runControl)
    {
        this.runControl = runControl;
    }

    public String parameterDescription() 
    {
        return "[timeBetweenChecks] [stabilityFactor]";
    }

    public String name()
    {
        return "wait_until_stable";
    }

    public boolean validWithProtocol (final String restriction)
    {
        return restriction.equals ("DT");
    }

    public String description() 
    {
        return "waits until a stable Delaunay Triangulation is formed.";
    }

    public String longDescription() 
    {
        return "Keeps checking until a stable Delaunay triangulation is formed by the nodes of the\n"+
            "sockets in the experiment.  A parameter can be used to adjust the time between checks.\n"+
            "The default time between checks is " + DEFAULT_STABLE_RESOLUTION + " seconds. Produces periodic progress updates\n" +
            "The next parameter is a stability factor that must be float in range [0,1], it determines what percentage\n" +
            "of sockets must be stable for the entire network to be considered to be stable.";
    }

    public int minParameters()
    {
        return 0;
    }

    public int maxParameters()
    {
        return 2;
    }

    public void action (final String          command,
                        final String          firstArg,
                        final String          secondArg,
                        final String          thirdArg,
                        final StringTokenizer furtherArgs,
                        final BufferedReader  in,
                        final PrintWriter     out,
                        final PrintWriter     err)
        throws Exception
    {
        if (WaitUntilStableCommand.debug)
            System.err.println ("WaitUntilStable action invoked");

        final long timeStartedWaiting = System.currentTimeMillis();
        int resolution = DEFAULT_STABLE_RESOLUTION;
        float stabilityFactor = DEFAULT_STABILITY_FACTOR;

        DT_WaitObject waitObject;

        synchronized (runControl.getSocketStatusTable().getLock()) 
        {
            
            if (firstArg != null)
            {
                resolution = toInt (firstArg);
                if (resolution <= 0)
                    throw new CommandFailureException ("Resolution must be positive: " + resolution);

                if (secondArg != null)
                {
                    stabilityFactor = toFloat (secondArg);
                    if (stabilityFactor < 0.0
                        ||
                        stabilityFactor > 1.0)
                    {
                        throw new CommandFailureException ("stabilityFactor must be float in range [0,1]: " + stabilityFactor);
                    }
                }
            }

            waitObject = new DT_WaitObject (runControl, out, err, stabilityFactor);
            waitObject.sendMessages();
        }
     
        // unlock the experiment but do not return from this command -
        // is this useful? correct? allow other commands to modify the
        // experiment while we're waiting for it to stabilize?

        while (! waitObject.isStable()) 
        {
            if (WaitUntilStableCommand.debug)
                System.err.println ("WaitUntilStable: looping");

            Thread.sleep (1000 * resolution); // do not catch InterruptedException, we want the user to have the ability to interrupt

            //synchronized (runControl.getExperimentLock()) 
            {
                
                //waitObject.update (experiment);
                waitObject.sendMessages();
            }
        }

        // debugging
        //System.err.println ("Setting experiment start time to: " + runControl.getExperiment().startTime + " now is: " + System.currentTimeMillis());

        final float t = (float) 0.001 * (System.currentTimeMillis() - timeStartedWaiting);
        final float u = (float) 0.001 * (System.currentTimeMillis() - runControl.getSocketStatusTable().getLastTimeStartingSockets());
        out.println ("\nStable at "+ u + " seconds since experiment started.   (Command took " + t + " seconds to run.)");
    }
 
    class DT_WaitObject 
    {
        RunControl runControl;
        //Experiment experiment;
        int nodesOnServers [];
        PrintWriter out;
        PrintWriter err;
     
        /** Complete GetValues request messages for Portals.  (As opposed
            to the "Replay" messages in <code>lightMsgs</code>. */
        MonMessage[] heavyMsgs;

        /** Addresses of the Portals in the experiment. */
        int[] addrs;

        /** Stores messages received. */
        MonMessage[] result = null;

        /** The number of runservers being used. */
        int msgsNeeded;

        /** The number of sockets being used. */
        int socketsUsed;

        /** The states of the sockets. */
        int states[];

        /** number between 0 and 1 that represents the fraction of
         * sockets that must be stable for the isStable method to
         * declare stability. For example if this is set to 0.75, then
         * isStable will return true if at least 3/4 of all sockets
         * report stable.  */
        private float stabilityFactor = DEFAULT_STABILITY_FACTOR;

        public DT_WaitObject (RunControl r, PrintWriter ps_out, PrintWriter ps_err, final float stabilityFactor) throws InterruptedException
        {
            this.stabilityFactor = stabilityFactor;
            runControl = r;
            out = ps_out;
            err = ps_err;
            
            nodesOnServers = r.getSocketStatusTable().getNodesOnServer();

            setNumOfMessages();
            makeMessages();
            remakeStateCounters();
        }

        
        /** Calculates the number of RunServers and the number of Sockets. */
        private void setNumOfMessages()
        {
            msgsNeeded = 0;
            socketsUsed = 0;
            for (int i = 0; i < nodesOnServers.length; i++)
            {
                if (nodesOnServers[i] > 0)
                {
                    msgsNeeded++;
                    socketsUsed += nodesOnServers[i];
                }
            }
        }

        /** Remakes the state counters for the experiment. */
        private void remakeStateCounters()
        {
            // used to store the LA and State of each node
            states = new int[socketsUsed];
            for (int i = 0; i < socketsUsed; i++)
            {
                states[i] = -1;
            }
        }
     
        /** Creates the cache of messages to be sent to the RunServers. */
        private void makeMessages()
        {
            I_Monitor monitor=runControl.getMonitor();
            heavyMsgs = new MonMessage[msgsNeeded];
            int j = 0;
            for (int i = 0; i < nodesOnServers.length; i++)
            {
                if (nodesOnServers[i] > 0)
                {
                    heavyMsgs[j] = monitor.createGetRequest();
                    for (int k = 1; k <= nodesOnServers[i]; k++)
                    {
                        heavyMsgs[j].addStat(XmlUtil.createXPath("/Portal/Appl/Socket["+k+"]/Node/State"));
                    }
                    j++;
                }
            }

            // get addresses messages
            addrs = new int[msgsNeeded];
            j = 0;
            for (int i = 0; i < nodesOnServers.length;i++)
            {
                if (nodesOnServers[i] > 0)
                {
                    addrs[j] = i;
                    j++;
                }
            }

            result = new MonMessage[msgsNeeded];
        }

        /**
         * Decides which RunServers should receive heavy and light messages, sends
         * the messages, and processes replies. 
         *
         * @throws InterruptedException
         * @throws CommandFailureException
         * @throws HyperCastStatsException
         * @throws NumberFormatException
         */
        public void sendMessages()
            throws InterruptedException, CommandFailureException, NumberFormatException, HyperCastStatsException
        {
            MonMessage[] msgsToSend = new MonMessage [msgsNeeded];

            // Figures out which messages to send...

            for (int i = 0; i < msgsNeeded; i++)
            {
                msgsToSend[i] = heavyMsgs[i];
            }

            // send messages

            CommandRunner cmdRunner = runControl.getCommandRunner();

            if (WaitUntilStableCommand.debug)
                System.err.println ("WaitUntilStable: sending queries");

            try
            {
                result = cmdRunner.sendQueries(addrs, msgsToSend); // blocks
            }
            catch (final NoSuchPortalException TODO)
            {
                TODO.printStackTrace();
                throw new CommandFailureException ("NoSuchPortalException caught", TODO);
            }

            if (WaitUntilStableCommand.debug)
                System.err.println ("WaitUntilStable: returned from send queries");

            // read out states of those that changed.
            final boolean ignoreErrors = true;
            int j = 0;
            int indexOffset = 0;
            for (int i = 0; i < nodesOnServers.length; i++)
            {
                if (nodesOnServers[i] > 0)
                {
                    if (result[j] == null)
                    {
                        if (ignoreErrors)
                        {
                            err.print("...(RunServer " + i + " did not report)");
                            err.flush();
                        }
                        else
                        {
                            err.println();
                            err.println("RunServer " + i + " did not report.");
                            err.println("Exiting command.");
                            return;
                        }
                    }
                    else
                    {
                        for (int k = 0; k < result[j].getStatsCount(); k++)
                        {
                            if(result[j].isStatError(k))
                            {
                                if (ignoreErrors)
                                {
                                    err.println("...(RunServer " + i + " encountered error: " + XmlUtil.getScalarValue(result[j].getStatValue(k)) + ")");
                                }
                                else
                                {
                                    err.println();
                                    err.println("RunServer " + i + " encountered error: " + XmlUtil.getScalarValue(result[j].getStatValue(k)) + ")");
                                    err.println("Exiting command.");
                                    return;
                                }
                            }
                            else
                            {
                                //Element result = (Element) nl.item(k).getChildNodes().item(1);
                                String rslt=XmlUtil.getScalarValue(result[j].getStatValue(k));
                                if (rslt.length() > 0) // TODO: include check that statistic is actually the state stat
                                {
                                    states[indexOffset+k] = Integer.parseInt(rslt);
                                }
                            }
                        }
                    }
                    // else "NACK" 
                    //  do nothing.

                    j++;
                    indexOffset += nodesOnServers[i];
                }
            }
        }

        /** Returns true if the topology is stable. */
        public boolean isStable()
        {
            int numStable = 0;
            int numLeader = 0;

            for (int i = 0; i < socketsUsed; i++)
            {
                if (states[i] == DT_Node_ServerJoin.STABLE
                    ||
                    states[i] == DT_Node_ServerJoin.LEADER_STABLE)
                {
                    numStable++;
                }

                if (states[i] == DT_Node_ServerJoin.LEADER_STABLE)
                {
                    numLeader++;
                }
            }

            if (numStable > stabilityFactor * socketsUsed)
            {
                int counter = 0;
                for (int i = 0; i < nodesOnServers.length; i++)
                {
                    for (int k = 0; k < nodesOnServers[i]; k++)
                    {
                        if (states[counter] != DT_Node_ServerJoin.STABLE
                            &&
                            states[counter] != DT_Node_ServerJoin.LEADER_STABLE)
                        {
                            out.println ("Waiting on " + i + "," + k);
                        }
                        counter++;
                    }
                }
            }

            if (numStable == socketsUsed
                &&
                numLeader == 1)
            {
                return true;
            }

            if (numStable != socketsUsed)
            {
                out.print ("..." + numStable);
            }
            else
            {
                out.print ("..." + numLeader + " Leaders");
            }

            out.flush();

            return false;
        }
    }
}
