/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

/* 
 * @author HyperCast Team
 */

// TODO Add comments 

package runserver;

import hypercast.ByteArrayUtility;
import hypercast.HyperCastConfig;
import hypercast.HyperCastStatsException;
import hypercast.I_LogicalAddress;
import hypercast.I_OverlayMessage;
import hypercast.I_OverlaySocket;
import hypercast.I_ReceiveCallback;
import hypercast.I_Stats;
import hypercast.Naming;
import hypercast.NotificationHandler;
import hypercast.OL_Socket;
import hypercast.SimpleStats;
import hypercast.StatsProcessor;
import hypercast.MonitorAndControl.I_Portal;
import hypercast.util.XmlUtil;

import java.util.Vector;

import org.apache.xpath.XPath;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

// Naming
import java.io.IOException;
import java.io.FileNotFoundException;
import java.util.StringTokenizer;
import java.io.FileInputStream;
import java.io.DataInputStream;
import java.io.ByteArrayInputStream;
import java.security.Security;
import java.security.NoSuchAlgorithmException;
import java.security.PrivateKey;
import java.security.KeyFactory;
import java.security.Provider;
import java.security.NoSuchProviderException;
import java.security.spec.PKCS8EncodedKeySpec;
import java.security.spec.InvalidKeySpecException;
import java.security.cert.Certificate;
import java.security.cert.CertificateException;
import java.security.cert.CertificateFactory;
import java.security.cert.X509Certificate;
import org.bouncycastle.jce.provider.BouncyCastleProvider;

/** 
 * This class manages an array of overlay sockets.  It allows remote
 * access to them through a Portal object, which translates XML
 * messages into I_Stats requests. Through those requests, the number
 * of sockets can be changed and the sockets can be started and
 * stopped.  Also, statistics of the sockets can be accessed.
 * 
 * @author Greg Mattes
 * @author Guangyu Dong
 * @author Jianping Wang
 * @author Michael Nahas
 * @version 3.0 November 2004
 */
public class RunServer
    implements I_Stats, Runnable
{
    protected static final boolean debug = false;

    /** Configuration shared by all of the sockets on this RunServer. */
    protected HyperCastConfig config = null;

    /**
     * A StatsProcessor instance which handles the statistics in this object.
     */
    private StatsProcessor statsProcessor;

    /**
     * Table of overlay sockets maintained in this object.
     */
    private SocketTable socketTable;

    // NFS file read block size in Version2 was usually 4k or 8k depending on
    // Client's page size. (source: Brian White)
    // Scott Ruffner reports that Version 3 is now the most popular version. (4/6/01)
    // NFS file read block size in Version 3 appears to usually be larger,
    // since TCP as well as UDP are being used for the tranport protocol.  
    // A bandwidth test of TCP vs. UDP vs. Blocksize is linked to below.  It
    // concludes that UDP with block sizes of 16k to 32k performed best.
    // See: http://www.netsys.com/sunmgr/1996-10/msg00100.html
    // I'll use 8k since it provides a smaller granularity.
    public static int FILE_BLOCK_SIZE = 8192;

    public static final int MIN_BYTES_PER_FILE_BLOCK = 9;

    public static int NUM_OF_PINGS = 30;
    public static int DELAY_BETWEEN_PINGS = 50; // in milliseconds

    // Arguments to constructor, need to save them so that they are available for run()
    private final int     initialSocketCount;
    private final boolean startSocketsAutomatically;
    private final String  portalConfigurationFilename;
    private final String  socketConfigurationFilename;
    private final String  configurationOverrides;

    protected I_Portal portal;


	/**
	 * The statistics name for this object (as an I_Stats instance)
	 */
	private String statisticsName = null;

 // Naming
    private static CertificateFactory certificateFactory;
    private static KeyFactory keyFactory;
    private static String certificateFormat = "X.509";
    private static String asymmetricEncryptionAlgorithm = "RSA";
    private static String cryptoBaseDir;

    /**
     * Creates an RunServer object and calls <code>go</code>.
     *
     * @see #usage()
     */
    public static void main (final String [] args)
    {
        int socketCount = 1;
        boolean startSocketsAutomatically = false;
        String portalConfigurationFilename = "RunServerSocket.xml";
        String socketConfigurationFilename = "hypercast.xml";
        final Vector configurationOverrides = new Vector();

        // Naming
        if (Security.getProvider ("BC") == null)
        {
            Security.insertProviderAt (new BouncyCastleProvider(), 2);
            //Security.addProvider (new BouncyCastleProvider());
        }

        // Naming
        try
        {
            certificateFactory = CertificateFactory.getInstance (certificateFormat, "BC");
            keyFactory = KeyFactory.getInstance (asymmetricEncryptionAlgorithm, "BC");
        }
        catch (final CertificateException ce)
        {
            System.err.println ("Could not instanciate CertificateFactory (" + certificateFormat + ")\n" + ce);
            System.exit (1);
        }
        catch (final NoSuchAlgorithmException nsae)
        {
            System.err.println ("Could not instanciate KeyFactory (" + asymmetricEncryptionAlgorithm + ")\n" + nsae);
            System.exit (1);
        }
        catch (final NoSuchProviderException nspe)
        {
            System.err.println ("Could not instanciate " + certificateFormat + " certificate factory, or " + asymmetricEncryptionAlgorithm + " key factory, provider BC unknown: " + nspe);
            System.exit (1);
        }

        for (int x = 0; x < args.length; ++x)
        {
            if (args[x].equals ("-ns"))
            {
                ++x;
                if (args.length < x+1)
                {
                    usage();
                }

                try
                {
                    socketCount = Integer.parseInt (args[x]);
                    if (socketCount <= 0)
                        throw new NumberFormatException();
                }
                catch (final NumberFormatException nfe)
                {
                    System.err.println ("Bad argument for number of socket (-ns): " + args[x]);
                    usage();
                }
            }
            else if (args[x].equals ("-start"))
            {
                startSocketsAutomatically = true;
            }
            else if (args[x].equals ("-nostart")) // this is the default
            {
                startSocketsAutomatically = false;
            }
            else if (args[x].equals ("-pc"))
            {
                ++x;
                if (args.length < x+1)
                {
                    usage();
                }
                portalConfigurationFilename = args[x];
            }
            else if (args[x].equals ("-sc"))
            {
                ++x;
                if (args.length < x+1)
                {
                    usage();
                }
                socketConfigurationFilename = args[x];
            }
            else if (args[x].startsWith ("-o")) // configuration override
            {
                ++x;
                if (args.length < x+1)
                {
                    usage();
                }
                configurationOverrides.add (args[x]);
            }
            // Naming
            else if (args[x].equals ("-cryptoDir"))
            {
                ++x;
                if (args.length < x+1)
                    usage();

                cryptoBaseDir = args [x];
            }
            // These options are superfluous since anything that is
            // not recognized does the same thing.
            else if (args[x].equals ("-h")
                     ||
                     args[x].equals ("-help"))
            {
                usage();
            }
            else
            {
                usage();
            }
        }

        final StringBuffer overrides = new StringBuffer();
        for (int i = 0; i < configurationOverrides.size(); ++i)
        {
            final String oneOverride = (String) configurationOverrides.get (i);
            overrides.append (oneOverride.trim());
            final boolean lastOne = (i == configurationOverrides.size() - 1);
            overrides.append (lastOne ? "" : "&");
        }

        final RunServer runServer = new RunServer (socketCount,
                                                   startSocketsAutomatically,
                                                   portalConfigurationFilename,
                                                   socketConfigurationFilename,
                                                   overrides.toString());

        new Thread (runServer, "Run Server").start();
    }

    /** Prints out the command-line arguments that are accepted by main(). */
    private static void usage()
    {
        final String usageString =
            "usage:"                                                                           + "\n" +
            "java runserver.RunServer [-ns <number of sockets>] [-nostart|-start]:"                      + "\n" +
            "               [-pc <portal config filename>] [-sc <socket config filename>]"     + "\n" +
            ""                                                                                 + "\n" +
            "-help:  Display this message"                                                     + "\n" +
            "-h:     Display this message"                                                     + "\n" +
            "-ns number of sockets:  The number of experimental sockets created by RunServer"  + "\n" +
            "-nostart | -start:      Whether or not to automatically start (call joinOverlay)" + "\n" +
            "                        at RunServer start time. (Default is nostart)"            + "\n" +
            "-pc portal config file: The config filename for portal's socket"                  + "\n" +
            "-sc socket config file: The config filename for experimental sockets"             + "\n";
            
        System.err.println (usageString);
        System.exit (1);
    }

    /**
     * Saves command line arguments for use in <code>run()</code>
     * method.
     */
    public RunServer (final int     initialSocketCount,
                      final boolean startSocketsAutomatically,
                      final String  portalConfigurationFilename,
                      final String  socketConfigurationFilename,
                      final String  configurationOverrides)
    {
        this.initialSocketCount = initialSocketCount;
        this.startSocketsAutomatically = startSocketsAutomatically;
        this.portalConfigurationFilename = portalConfigurationFilename;
        this.socketConfigurationFilename = socketConfigurationFilename;
        this.configurationOverrides = configurationOverrides;

		this.setStatsName("Appl");

        createOverlayAndConfig (this.socketConfigurationFilename, this.configurationOverrides);

		statsProcessor = new StatsProcessor(this/* I_Stats object */, true/* readable flag */, true/* writable flag */);

        this.socketTable = new SocketTable (this.initialSocketCount, this.config, this.statsProcessor);

        /*
         * Socket Table must be created before statistics are
         * initialized; config must be created before Socket Table.
         */
        initStatisticsStructure();
    }

    public void run()
    {
        if (this.startSocketsAutomatically)
        {
            this.socketTable.startAllSockets();
        }

        /*
         * After everything is created, read Portal configuration and
         * open the Portal; there is no need to keep a reference to
         * the Portal. All commands that control the RunServer will be
         * received via the Portal. These commands are translated into
         * statistical queries that get and set values. The Portal
         * will call "getStats" or "setStats" which will in turn call
         * the statsProcessor of this class which will either delegate
         * the call to SocketTable (see initStatisticsStructure) or it
         * will call "getStatsValue" or "setStatsValue."
         */
        final HyperCastConfig portalConfig  = HyperCastConfig.createConfig (portalConfigurationFilename);
        //this.portal = portalConfig.createPortal (this, "Appl");
		this.portal = portalConfig.createPortal (this, statisticsName);
        this.portal.openPortal();
    }

    /**
     * Creates the Overlay and the HyperCastConfiguration
     * object.
     */
    private void createOverlayAndConfig (final String configurationFilename, final String configurationOverrides)
    {
    	
    	config = HyperCastConfig.createConfig(configurationFilename);
    	
    	/*
    	 * Set private attribute to run security setting
    	 */
    	config.setPrivateTextAttribute(XmlUtil.createXPath("/Private/KeyStorePassword"), "password");
    	config.setPrivateTextAttribute(XmlUtil.createXPath("/Private/PrivateKeyAlias"), "testpair");
    	config.setPrivateTextAttribute(XmlUtil.createXPath("/Private/PrivateKeyPassword"), "password");
    	config.setPrivateTextAttribute(XmlUtil.createXPath("/Private/GroupKey"), "1234567812345678");
    	
    }

    private void initStatisticsStructure()
    {
        //statsProcessor = new StatsProcessor(this/* I_Stats object */, true/* readable flag */, true/* writable flag */);

        statsProcessor.addStatsElement ("NumOfSockets"/* name */, new SocketNumQuery()/* I_Stats object */, 1/* minoccurs */, 1/* maxoccurs */);

        statsProcessor.addStatsElement ("CurrentTime", new CurrentTime(), 1, 1);

        statsProcessor.addStatsElement ("KillApp", new KillApp(), 1, 1);

        statsProcessor.addStatsElement ("SetName", new SetName(), 1, 1);
    }

	/**
	 * This class manages simple statistic "KillApp".
	 */
	class KillApp extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return null;
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			if (newValue.equals ("true"))
			{
				// NB: no response sent to sender

				try{
					System.out.println("Closing Portal...");
					portal.closePortal();
				}finally{
					System.exit (0);
				}
			}
			
			return newValue;
		}

		/** 
		 * @see I_Stats#getQriteSchame
		 */
		public Element[] getWriteSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Boolean"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}

	/**
	 * This class manages simple statistic "KillApp".
	 */
	class SetName extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return null;
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
                    String nameString = null;
                    try
                    {
                        final String cryptoDir = cryptoBaseDir + "/grm/crypto";

                        final String certificateFilenameCA = cryptoDir + "/cert-CA-HC.der";

                        //System.err.println ("Reading CA certificate: " + certificateFilenameCA);

                        final Certificate certificateCA = readX509Certificate (certificateFilenameCA, certificateFactory);
                
                        //System.err.println ("RunServer SetName: " + newValue);

                        // format for individually addressed socket: <socket index>:<name>
                        // only one allowed per set, i.e. no comma separated names
                        if (newValue.indexOf (":") != -1)
                        {
                            final StringTokenizer st = new StringTokenizer (newValue, ":");

                            if (st.countTokens() != 2)
                                throw new HyperCastStatsException ("Format of set name for individually addressed socket is <socket index>:<name>: " + newValue, HyperCastStatsException.IncorrectSetValue);

                            try
                            {
                                final String indexString = st.nextToken();
                                final int socketIndex = Integer.parseInt (indexString);
                                nameString = st.nextToken();

                                if (nameString.equals ("9999") || true)
                                {
                                    final String certificateFilename = cryptoDir + "/" + nameString.substring (0,2) + "/cert-" + nameString + ".der";
                                    final String privateKeyFilename = cryptoDir + "/" + nameString.substring (0,2) + "/privkey-" + nameString + ".der";

                                    final Certificate certificate = readX509Certificate (certificateFilename, certificateFactory);
                                    final PrivateKey privateKey = readPrivateKey (privateKeyFilename, keyFactory);

                                    System.err.println ("RunServer SetName index: " + socketIndex + " string: " + nameString +
                                                        " cert filename: " + certificateFilename + " key filename: " + privateKeyFilename +
                                                        " cert: " + ((certificate == null) ? "null" : "not null") +
                                                        " key: " + ((privateKey == null) ? "null" : "not null"));

                                    socketTable.get (socketIndex).installTrustedNamingCertificate (certificateCA);
                                    socketTable.get (socketIndex).setName (certificate, privateKey);

                                    return newValue;
                                }
                                else
                                {
                                    socketTable.get (socketIndex).installTrustedNamingCertificate (certificateCA);

                                    for (int n = 0; n < 300; ++n)
                                    {
                                        final int length = 4;

                                        final String orig = Integer.toString (n);
                                        final int numOfLeadingZeros = length - orig.length();
                                        if (numOfLeadingZeros < 0) { throw new RuntimeException(); }
                                        final StringBuffer stringBuffer = new StringBuffer (length);
                                        for (int j = 0; j < length; ++j) { stringBuffer.append ('0'); }
                                        for (int j = numOfLeadingZeros; j < length; ++j) { stringBuffer.setCharAt (j, orig.charAt (j - numOfLeadingZeros)); }
                                        nameString = stringBuffer.toString();

                                        final String certificateFilename = cryptoDir + "/" + nameString.substring (0,2) + "/cert-" + nameString + ".der";
                                        final String privateKeyFilename = cryptoDir + "/" + nameString.substring (0,2) + "/privkey-" + nameString + ".der";

                                        final Certificate certificate = readX509Certificate (certificateFilename, certificateFactory);
                                        final PrivateKey privateKey = readPrivateKey (privateKeyFilename, keyFactory);

/*
                                        System.err.println ("RunServer SetName index: " + socketIndex + " string: " + nameString +
                                                            " cert filename: " + certificateFilename + " key filename: " + privateKeyFilename +
                                                            " cert: " + ((certificate == null) ? "null" : "not null") +
                                                            " key: " + ((privateKey == null) ? "null" : "not null"));
*/

                                        socketTable.get (socketIndex).setName (certificate, privateKey);
                                    }
                                    return newValue;
                                }
                            }
                            catch (final NumberFormatException nfe)
                            {
                                throw new HyperCastStatsException ("Format of set name for individually addressed socket is <socket index>:<name>: " + nfe, HyperCastStatsException.IncorrectSetValue);
                            }
                        }

                        final StringTokenizer st = new StringTokenizer (newValue, ",");
            
                        int socketIndex = 0;
                        while (st.hasMoreTokens())
                        {
                            nameString = st.nextToken();

                            final String certificateFilename = cryptoDir + "/" + nameString.substring (0,2) + "/cert-" + nameString + ".der";
                            final String privateKeyFilename = cryptoDir + "/" + nameString.substring (0,2) + "/privkey-" + nameString + ".der";

                            final Certificate certificate = readX509Certificate (certificateFilename, certificateFactory);
                            final PrivateKey privateKey = readPrivateKey (privateKeyFilename, keyFactory);

/*
                            System.err.println ("RunServer SetName index: " + socketIndex + " string: " + nameString +
                                                " cert filename: " + certificateFilename + " key filename: " + privateKeyFilename +
                                                " cert: " + ((certificate == null) ? "null" : "not null") +
                                                " key: " + ((privateKey == null) ? "null" : "not null"));
*/

                            socketTable.get (socketIndex).installTrustedNamingCertificate (certificateCA);
                            socketTable.get (socketIndex).setName (certificate, privateKey);
                            ++socketIndex;
                        }
                        return "RunServer SetName: " + newValue;
                    }
                    catch (final Exception e)
                    {
                        final String errorString = "Exception caught while setting name (" + nameString + ", All names: " + newValue + "): " + e;
                        System.err.println (errorString);
                        throw new HyperCastStatsException (errorString, HyperCastStatsException.InappropriateValue);
                    }
		}

		/** 
		 * @see I_Stats#getQriteSchame
		 */
		public Element[] getWriteSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:String"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}        

	/**
	 * This class manages simple statistic "NumOfSockets".
	 */
	class SocketNumQuery extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return Integer.toString (socketTable.size());
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			socketTable.setSocketCount (Integer.parseInt (newValue));
			return newValue;
		}

		/** 
		 * @see I_Stats#getReadSchame
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Integer"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}

		/** 
		 * @see I_Stats#getQriteSchame
		 */
		public Element[] getWriteSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Integer"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}


	/**
	 * This class manages simple statistic "CurrentTime".
	 */
	class CurrentTime extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return Long.toString (System.currentTimeMillis());
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}

		/** 
		 * @see I_Stats#getReadSchame
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Long"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}

    /** @see I_Stats#getStats */
    public Element[] getStats (final Document doc, final XPath xpath)
        throws HyperCastStatsException
    {
        return statsProcessor.getStatsResult (doc, xpath);
    }

    /** @see I_Stats#setStats */
    public Element[] setStats (final Document doc, final XPath xpath, final Element newValue)
        throws HyperCastStatsException
    {
        return statsProcessor.setStatsResult (doc, xpath, newValue);
    }

    /** @see I_Stats#getReadSchema */
    public Element[] getReadSchema (final Document doc, final XPath xpath)
        throws HyperCastStatsException
    {
        return statsProcessor.getReadSchemaResult (doc, xpath);
    }

    /** @see I_Stats#getWriteSchema */
    public Element[] getWriteSchema (final Document doc, final XPath xpath)
        throws HyperCastStatsException
    {
        return statsProcessor.getWriteSchemaResult (doc, xpath);
    }


	/** @see I_Statis#getStatsName */
	public String getStatsName ()
	{
		return statisticsName;
	}

	/** @see I_Statis#setStatsName */
	public void setStatsName (String name)
	{
		statisticsName = name;
	}

    // rather than creating a new object for each socket, use this one many times
    protected static final I_ReceiveCallback nullCallback = new NullCallback();

    /**
     * This class manages the array of overlay sockets.
     */
    class SocketTable
     {
        private I_OverlaySocket [] socketTable = new I_OverlaySocket [0]; // could/should be ArrayList

        private HyperCastConfig config;
        
		StatsProcessor statsProcessor = null;
		
		String statsName = "Socket";

        public SocketTable (final int initialSocketCount, final HyperCastConfig config, StatsProcessor sProcessor)
        {
        	this.config = config;

			statsProcessor = sProcessor;

            setSocketCount (initialSocketCount);
        }

        // Naming
        public synchronized I_OverlaySocket get (final int i) {
            return socketTable [i];
        }

        public synchronized int size()
        {
            return socketTable.length;
        }

        public synchronized void startAllSockets()
        {
            for (int i = 0; i < socketTable.length; ++i)
            {
                socketTable [i].joinOverlay(); // no-op if already running
            }
        }

        /**
         * This is the utility function that makes these two features work:
         * DelayStartSocketSeries
         * DelayStopSocketSeries
         */
        public synchronized void setRunningSocketSeries (final int startSocket, final int endSocket, final boolean setRunningTo)
        {
            if (startSocket < 0
                ||
                endSocket >= size())
            {
                throw new IndexOutOfBoundsException ("Range [" + startSocket + ", " + endSocket + "] is out of range for SocketTable with " + size() + " entries");
            }

            if (setRunningTo)
            {
                for (int i = startSocket; i <= endSocket; ++i)
                {
                    socketTable[i].joinOverlay();
                }
            }
            else
            {
                for (int i = startSocket; i <= endSocket; ++i)
                {
                    socketTable[i].leaveOverlay();
                }
            }
        }

        /**
         * Could be smaller or larger than current count, must handle both cases
         */
        public synchronized void setSocketCount (final int newSize)
        {
            if (RunServer.debug)
                System.err.println ("setSocketCount: " + newSize);

            if (newSize < socketTable.length)
            {
                final I_OverlaySocket [] smallerSocketTable = new I_OverlaySocket [newSize];
                System.arraycopy (socketTable, 0, smallerSocketTable, 0, newSize);

                for (int i = (socketTable.length - 1); i >= newSize; --i)
                {
                    //statsProcessor.removeStatsElement (statsName, i);
					statsProcessor.removeStatsElement (socketTable[i]);
                    socketTable [i].leaveOverlay(); // no-op if already not running
                }
                socketTable = smallerSocketTable;
            }
            else if (newSize > socketTable.length)
            {
                final I_OverlaySocket [] largerSocketTable = new I_OverlaySocket [newSize];
                System.arraycopy (socketTable, 0, largerSocketTable, 0, socketTable.length);

                for (int i = socketTable.length; i < newSize; ++i)
                {
                    final I_OverlaySocket newSocket = createSocket (new SilentNullCallback());
                    //final I_OverlaySocket newSocket = createSocket();
                    largerSocketTable [i] = newSocket;
                    
                    addNewTableEntryToStatsProcessor (newSocket);
                }
                socketTable = largerSocketTable;
            }
            else // (newSize == socketTable.length)
            {
                // do nothing
            }
        }

        /**
         * Creates a new Socket. A <code>NotificationHandler</code> is
         * needed so that some features, such as naming, will work.
         *
         * @see OL_Socket#OL_Socket(HyperCastConfig,I_ReceiveCallback,NotificationHandler)
         * @see Naming#Naming(OL_Socket)
         */
        private I_OverlaySocket createSocket (final I_ReceiveCallback callback)
        {
            return config.createOverlaySocket (callback, new NotificationHandler());
        }

        private I_OverlaySocket createSocket()
        {
            return config.createOverlaySocket (RunServer.this.nullCallback, new NotificationHandler());
        }

        

        private void addNewTableEntryToStatsProcessor (final I_OverlaySocket socket)
        {
			statsProcessor.addStatsElement ("Socket", socket, 0, -1);
        	
        }

        private void removeAllTableEntryInStatsProcessor()
        {
            statsProcessor.removeAllStatsElement(statsName);
        	//clear();
        }
    }

    // Naming
    private static PrivateKey readPrivateKey (String privateKeyFilename, KeyFactory kf)
    {
        PrivateKey privateKey = null;

        try
        {
            FileInputStream fis = new FileInputStream (privateKeyFilename);
            DataInputStream dis = new DataInputStream (fis);
            byte [] privateKeyData = new byte [dis.available()];
            dis.readFully (privateKeyData);
            //System.out.println ("Private Key Data Size: " + privateKeyData.length);
            dis.close();

            // Instanciate private key object using data from file
            PKCS8EncodedKeySpec privateKeySpec = new PKCS8EncodedKeySpec (privateKeyData);
            privateKey = kf.generatePrivate (privateKeySpec);

            //System.out.println ("format: " + privateKeySpec.getFormat() + "\n" + privateKey);
        }
        catch (final FileNotFoundException fnfe)
        {
            System.err.println (fnfe);
        }
        catch (final IOException ioe)
        {
            System.err.println (ioe);
        }
        catch (final InvalidKeySpecException ikse)
        {
            System.err.println (ikse);
        }


        return privateKey;
    }

    // Naming
    private static X509Certificate readX509Certificate (String certificateFilename, CertificateFactory cf)
    {
        X509Certificate certificate = null;

        try
        {
            // Read certificate data from file

            FileInputStream fis = new FileInputStream (certificateFilename);
            DataInputStream dis = new DataInputStream (fis);
            byte [] certificateData = new byte [dis.available()];
            dis.readFully (certificateData);
            //System.out.println ("Certificate Data Size: " + certificateData.length);
            dis.close();

            // Instanciate certificate object using data from file

            ByteArrayInputStream bais = new ByteArrayInputStream (certificateData);
            certificate = (X509Certificate) cf.generateCertificate (bais);
        }
        catch (final FileNotFoundException fnfe)
        {
            System.err.println (fnfe);
        }
        catch (final IOException ioe)
        {
            System.err.println (ioe);
        }
        catch (final CertificateException ce)
        {
            System.err.println (ce);
        }

        return certificate;
    }
}

/**
 * A callback used when none is needed. The messageArrived method
 * throws an exception when called since it should never be called
 * (but you would sure like to know if it is ever called since that
 * would indicate bug).
 */
class NullCallback
    implements I_ReceiveCallback
{
    public void ReceiveCallback (final I_OverlayMessage message)
    {
        throw new UnsupportedOperationException ("Message was received in NullCallback\n" + message);
    }
}

class SilentNullCallback
    implements I_ReceiveCallback
{
    public void ReceiveCallback (final I_OverlayMessage ignored) { }
}

/**
 * A simple callback for the sockets. 
 */
class RunServerCallback
    implements I_ReceiveCallback
{
    public static byte PING = (byte) 1;
    public static byte PONG = (byte) 2;
    public static byte FILE_BLOCK = (byte) 3;

    private static int PING_MSG_SIZE = 1 + 4 + 4 + 4;

    I_OverlaySocket socket;

    // Vector containing Vectors containing successful roundtrip measurements using PING/PONG
    public Vector roundTripTimes;

    // Vector containing Vectors containing initial send times of all roundtrip measurements
    private Vector pingSentTimes;

    //Vectors containing hop count of round trip of PING/PONG
    private Vector hopCountsOutgoing;
    private Vector hopCountsIncoming;

    //Vector containing Vectors of times when BW Test messages were sent.
    // The time at index 0 is the time the test started.
    // The time at index N is the time the Nth message completed sending.
    private Vector fileBlockSendTimes;

    //Vector containing Vectors of times when BW Test messages were received.
    private Vector fileBlockReceiveTimes;

    private Vector duplicateBlocks;

    private Vector lastReceiveTime;
    public RunServerCallback() {
        pingSentTimes = new Vector();
        roundTripTimes = new Vector();
        hopCountsOutgoing = new Vector();
        hopCountsIncoming = new Vector();

        fileBlockSendTimes = new Vector();
        fileBlockReceiveTimes = new Vector();
        duplicateBlocks = new Vector();
        lastReceiveTime = new Vector();

    }

    public void setSocket(I_OverlaySocket s) {
        socket = s;
    }

    public void ReceiveCallback(I_OverlayMessage msg) {
        byte[] payload = msg.getPayload();
        if (msg.getPayloadLength() == PING_MSG_SIZE && (payload[0] == PING || payload[0] == PONG))
            handlePing(msg);
        else if (msg.getPayloadLength() > 0 && payload[0] == FILE_BLOCK)
            handleFileBlock(msg);
        else
            System.err.println("RunServerCallback received overlay message with unknown payload.");
    }

    public void sendPing(I_LogicalAddress target, int measurementIndex, int attempt) {
        byte[] payload = new byte[PING_MSG_SIZE];
        payload[0] = PING;
        System.arraycopy(ByteArrayUtility.toByteArray(measurementIndex), 0, payload, 1, 4);
        System.arraycopy(ByteArrayUtility.toByteArray(attempt), 0, payload, 1 + 4, 4);

        I_OverlayMessage msg = socket.createMessage(payload);

        if (measurementIndex >= pingSentTimes.size())
            pingSentTimes.setSize(measurementIndex + 1);

        Vector v = (Vector) pingSentTimes.elementAt(measurementIndex);
        if (v == null) {
            v = new Vector();
            pingSentTimes.setElementAt(v, measurementIndex);
        }

        if (v.size() <= attempt) {
            v.setSize(attempt + 1);
        }

        Long timeSent = new Long(System.currentTimeMillis());
        v.setElementAt(timeSent, attempt);
        socket.sendToNode(msg, target);
    }

    private void handlePing(I_OverlayMessage msg) {
        byte[] payload = msg.getPayload();
        if (payload[0] == PING) {
            // send PONG
            byte[] payload2 = new byte[PING_MSG_SIZE];
            payload2[0] = PONG;
            System.arraycopy(payload, 1, payload2, 1, 8);

            int hopcount = socket.getHopLimit() - msg.getHopLimit();
            System.arraycopy(ByteArrayUtility.toByteArray(hopcount), 0, payload2, 9, 4);

            I_OverlayMessage msg2 = socket.createMessage(payload2);
            socket.sendToNode(msg2, msg.getSourceAddress());
        } else if (payload[0] == PONG) {
            // record round-trip time
            long timeReceived = System.currentTimeMillis();
            int measurementIndex = ByteArrayUtility.toInteger(payload, 1);
            int attempt = ByteArrayUtility.toInteger(payload, 1 + 4);
            int outgoingHopCount = ByteArrayUtility.toInteger(payload, 1 + 4 + 4);
            if (measurementIndex < pingSentTimes.size()
                && pingSentTimes.elementAt(measurementIndex) != null) {
                Vector v = (Vector) pingSentTimes.elementAt(measurementIndex);
                if (attempt < v.size() && v.elementAt(attempt) != null) {
                    Long timeSent = (Long) v.elementAt(attempt);
                    long rtt = timeReceived - timeSent.longValue();
                    Vector times;
                    if (roundTripTimes.size() <= measurementIndex) {
                        times = new Vector();
                        roundTripTimes.setSize(measurementIndex + 1);
                        roundTripTimes.setElementAt(times, measurementIndex);
                    } else {
                        times = (Vector) roundTripTimes.elementAt(measurementIndex);
                        if (times == null) {
                            times = new Vector();
                            roundTripTimes.setElementAt(times, measurementIndex);
                        }
                    }

                    if (times.size() <= attempt) {
                        times.setSize(attempt + 1);
                    }

                    times.setElementAt(new Long(rtt), attempt);

                    // calculate HopCount
                    int incomingHopCount = (socket.getHopLimit() - msg.getHopLimit());
                    if (hopCountsOutgoing.size() <= measurementIndex) {
                        hopCountsOutgoing.setSize(measurementIndex + 1);
                        hopCountsIncoming.setSize(measurementIndex + 1);
                    }

                    hopCountsOutgoing.setElementAt(new Integer(outgoingHopCount), measurementIndex);
                    hopCountsIncoming.setElementAt(new Integer(incomingHopCount), measurementIndex);
                } else {
                    System.err.println(
                        "PingResponder received PONG #"
                        + attempt
                        + " when PING #"
                        + attempt
                        + "  had not been sent!");
                }
            } else {
                System.err.println("PingResponder received PONG from node that was not sent PING!");
            }
        }
    }

    public double getAvgRTT(int measurementIndex) {
        if (roundTripTimes.size() <= measurementIndex)
            return -1;

        Vector v = (Vector) roundTripTimes.elementAt(measurementIndex);

        if (v == null)
            return -1;

        long sum = 0;
        int count = 0;
        for (int i = 0; i < v.size(); i++) {
            if (v.elementAt(i) != null) {
                long rtt = ((Long) v.elementAt(i)).longValue();
                sum += rtt;
                count++;
            }
        }

        if (count == 0)
            return -1;
        else
            return (sum / 1000.0) / (double) count;
    }

    public double getRTTStandardDeviation(int measurementIndex) {
        double avg = getAvgRTT(measurementIndex);
        if (avg < 0)
            return -1;

        Vector v = (Vector) roundTripTimes.elementAt(measurementIndex);

        double sum = 0;
        int count = 0;
        for (int i = 0; i < v.size(); i++) {
            if (v.elementAt(i) != null) {
                long rtt = ((Long) v.elementAt(i)).longValue();
                double diff = (rtt / 1000.0) - avg;
                sum += diff * diff;
                count++;
            }
        }

        return Math.sqrt(sum / count);
    }

    public int getNumOfSuccessfulPings(int measurementIndex) {
        if (roundTripTimes.size() <= measurementIndex)
            return 0;

        Vector v = (Vector) roundTripTimes.elementAt(measurementIndex);

        if (v == null)
            return 0;

        int count = 0;
        for (int i = 0; i < v.size(); i++) {
            if (v.elementAt(i) != null)
                count++;
        }

        return count;
    }

    public int getHopCountOutgoing(int measurementIndex) {
        if (measurementIndex < hopCountsOutgoing.size()) {
            Integer hopCount = (Integer) hopCountsOutgoing.elementAt(measurementIndex);
            if (hopCount != null)
                return hopCount.intValue();
            else
                return -1;
        } else {
            return -1;
        }
    }

    public int getHopCountIncoming(int measurementIndex) {
        if (measurementIndex < hopCountsIncoming.size()) {
            Integer hopCount = (Integer) hopCountsIncoming.elementAt(measurementIndex);
            if (hopCount != null)
                return hopCount.intValue();
            else
                return -1;
        } else {
            return -1;
        }
    }

//TODO:
    public int getMeasureNum()
    {
        return roundTripTimes.size();
    }

    public void multicastFileBlock(int blockSize, int measurementIndex) {
        if (fileBlockSendTimes.size() <= measurementIndex) {
            fileBlockSendTimes.setSize(measurementIndex + 1);
        }

        Vector v = (Vector) fileBlockSendTimes.elementAt(measurementIndex);

        if (v == null) {
            v = new Vector();
            fileBlockSendTimes.setElementAt(v, measurementIndex);
            v.addElement(new Long(System.currentTimeMillis()));
        }

        int blockNumber = v.size() - 1;

        blockSize = Math.max(RunServer.MIN_BYTES_PER_FILE_BLOCK, blockSize);
        byte[] payload = new byte[blockSize];
        payload[0] = FILE_BLOCK;
        System.arraycopy(ByteArrayUtility.toByteArray(measurementIndex), 0, payload, 1, 4);
        System.arraycopy(ByteArrayUtility.toByteArray(blockNumber), 0, payload, 1 + 4, 4);
        I_OverlayMessage msg = socket.createMessage(payload);

        socket.sendToAll(msg);
        v.addElement(new Long(System.currentTimeMillis()));
    }

    public void handleFileBlock(I_OverlayMessage msg) {
        byte[] payload = msg.getPayload();
        long timeReceived = System.currentTimeMillis();
        int measurementIndex = ByteArrayUtility.toInteger(payload, 1);
        int blockNumber = ByteArrayUtility.toInteger(payload, 1 + 4);

        if (fileBlockReceiveTimes.size() <= measurementIndex) {
            fileBlockReceiveTimes.setSize(measurementIndex + 1);
        }
        if (duplicateBlocks.size() <= measurementIndex) {
            duplicateBlocks.setSize(measurementIndex + 1);
        }

        Vector times = (Vector) fileBlockReceiveTimes.elementAt(measurementIndex);
        if (times == null) {
            times = new Vector();
            fileBlockReceiveTimes.setElementAt(times, measurementIndex);
        }

        if (times.size() <= blockNumber) {
            times.setSize(blockNumber + 1);
        }

        if (times.elementAt(blockNumber) != null) {
            Long duplicates = (Long) duplicateBlocks.elementAt(measurementIndex);
            if (duplicates == null) {
                duplicateBlocks.setElementAt(new Long(1), measurementIndex);
            } else {
                long t = duplicates.longValue();
                t++;
                duplicateBlocks.setElementAt(new Long(t), measurementIndex);
            }
        } else {
            if (lastReceiveTime.size() <= measurementIndex) {
                lastReceiveTime.setSize(measurementIndex + 1);
            }
            lastReceiveTime.setElementAt(new Long(System.currentTimeMillis()), measurementIndex);

        }
        times.setElementAt(new Long(System.currentTimeMillis()), blockNumber);

    }

    public long getStartTimeOfBWTest(int measurementIndex) {
        if (fileBlockSendTimes.size() <= measurementIndex)
            return -1;

        Vector v = (Vector) fileBlockSendTimes.elementAt(measurementIndex);
        if (v == null)
            return -1;

        return ((Long) v.elementAt(0)).longValue();
    }

    public int getNumOfArrivals(int measurementIndex) {
        if (fileBlockReceiveTimes.size() <= measurementIndex)
            return -1;

        Vector v = (Vector) fileBlockReceiveTimes.elementAt(measurementIndex);
        if (v == null)
            return -1;

        int count = 0;
        for (int i = 0; i < v.size(); i++) {
            if (v.elementAt(i) != null)
                count++;
        }

        return count;
    }

    public long getSendTimeOfNthBlock(int measurementIndex, int blockIndex) {
        if (fileBlockSendTimes.size() <= measurementIndex)
            return -1;

        Vector v = (Vector) fileBlockSendTimes.elementAt(measurementIndex);
        if (v == null || v.size() <= (blockIndex + 1))
            return -1;

        return ((Long) v.elementAt(blockIndex + 1)).longValue();
    }

    // NB: This returns the average of the time before
    // the send function was called.  This is to prevent
    // problems where the avg. receive time could be 
    // less than the avg. send time.
    public long getAvgSendTimeOfBlocks(int measurementIndex) {
        if (fileBlockSendTimes.size() <= measurementIndex)
            return -1;

        Vector v = (Vector) fileBlockSendTimes.elementAt(measurementIndex);
        if (v == null)
            return -1;

        if (v.elementAt(0) == null)
            return 0;

        long sendTimeOfFirstMsg = ((Long) v.elementAt(0)).longValue();

        // Following loop calcuates avg. delay after 1st msg.
        // The delay of the first message will be added later.
        // This is done to prevent "sum" from overflowing.
        long sum = 0;
        int count = 0;
        for (int i = 0; i < v.size() - 1; i++) {
            Long sendTimeOfIthMsg = (Long) v.elementAt(i);
            if (sendTimeOfIthMsg != null) {
                sum += (sendTimeOfIthMsg.longValue() - sendTimeOfFirstMsg);
                count++;
            }
        }

        return sendTimeOfFirstMsg + (sum / count);
    }

    public long getReceiveTimeOfNthBlock(int measurementIndex, int blockIndex) {
        if (fileBlockReceiveTimes.size() <= measurementIndex)
            return -1;

        Vector v = (Vector) fileBlockReceiveTimes.elementAt(measurementIndex);
        if (v == null || v.size() <= blockIndex)
            return -1;

        if (v.elementAt(blockIndex) == null)
            return -1;
        else
            return ((Long) v.elementAt(blockIndex)).longValue();
    }

    /**Added by Guimin
     */
    public long getReceiveTimeOfLastBlock(int measurementIndex) {
        if (lastReceiveTime.size() <= measurementIndex)
            return -1;
        else
            return ((Long) (lastReceiveTime.elementAt(measurementIndex))).longValue();

    }

    public long getNumOfDuplicates(int measurementIndex) {
        if (duplicateBlocks.size() <= measurementIndex)
            return 0;
        else {
            Long db = (Long) (duplicateBlocks.elementAt(measurementIndex));
            if (db == null)
                return 0;
            else
                return db.longValue();
        }
    }
    public long getAvgReceiveTimeOfBlocks(int measurementIndex) {
        if (fileBlockReceiveTimes.size() <= measurementIndex)
            return -1;

        Vector v = (Vector) fileBlockReceiveTimes.elementAt(measurementIndex);
        if (v == null)
            return -1;

        if (v.elementAt(0) == null)
            return -1;

        long receiveTimeOfFirstMsg = ((Long) v.elementAt(0)).longValue();

        // Following loop calcuates avg. delay after 1st msg.
        // Later, we'll add in delay of 1st msg.
        // This is done to prevent "sum" from overflowing.
        long sum = 0;
        int count = 0;
        for (int i = 0; i < v.size(); i++) {
            Long receiveTimeOfIthMsg = (Long) v.elementAt(i);
            if (receiveTimeOfIthMsg != null) {
                sum += (receiveTimeOfIthMsg.longValue() - receiveTimeOfFirstMsg);
                count++;
            }
        }

        return receiveTimeOfFirstMsg + (sum / count);
    }
}

class OverlayPingSender implements Runnable {
    public static int MSG_SIZE = 9;

    private I_LogicalAddress target;
    private RunServerCallback socketManager;
    private int measurementIndex;
    public OverlayPingSender(I_LogicalAddress addr, RunServerCallback s, int mi) {
        target = addr;
        socketManager = s;
        measurementIndex = mi;
    }

    public void run() {
        for (int attempt = 0; attempt < RunServer.NUM_OF_PINGS; attempt++) {
            try {
                socketManager.sendPing(target, measurementIndex, attempt);
                Thread.sleep(RunServer.DELAY_BETWEEN_PINGS);
            } catch (Exception e) {
                System.err.println("Exception occured while sending out PINGs: " + e.toString());
            }
        }
    }
}

class OverlayBWTestSender implements Runnable {
    private int numberOfBlocks;
    private int blockSize;
    private RunServerCallback socketManager;
    private int measurementIndex;

    private int rate;

    public OverlayBWTestSender(int n, int size, RunServerCallback s, int mi) {
        numberOfBlocks = n;
        blockSize = size;
        socketManager = s;
        measurementIndex = mi;
    }

    public OverlayBWTestSender(int n, int size, int sendRate, RunServerCallback s, int mi) {
        numberOfBlocks = n;
        blockSize = size;
        socketManager = s;
        measurementIndex = mi;
        rate = sendRate;
    }

    public void run() {
        long fix_sleep_time;
        long start_time = 0;
        long end_time = 0;
        int pkgs_onetime = 3;

        if (rate == 0) {
            fix_sleep_time = 0;
        } else {
            fix_sleep_time = (new Float((blockSize * 8 * pkgs_onetime) / rate)).longValue();
            //System.out.println("fix_sleep_time=" + fix_sleep_time);
        }

        start_time = System.currentTimeMillis();
        for (int block = 0; block < numberOfBlocks; block++) {
            try {
                socketManager.multicastFileBlock(blockSize, measurementIndex);

                end_time = System.currentTimeMillis();

                //if (rate > 0)
                if ((rate > 0) && (block % pkgs_onetime == 0) && (block > 0)) {
                    long sleep_time = fix_sleep_time - end_time + start_time;
                    //System.out.println("sleep time=" + sleep_time);
                    if (sleep_time > 0) {
                        try {
                            //Thread.sleep(sleepTimeByRate(rate, blocksize));
                            //Thread.sleep(new Float((blockSize*8)/rate).longValue());
                            Thread.sleep(sleep_time > 10 ? sleep_time - 10 : sleep_time);
                        } catch (InterruptedException e) {
                        }
                    } else {
                        System.out.println("The specified sending rate is too fast!");
                    }
                }
                start_time = System.currentTimeMillis();

            } catch (Exception e) {
                System.err.println(
                    "Exception occured while sending out FILE BLOCK: " + e.toString());
            }
        }
    }
}

